/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

#include "stdafx.h"
#include "ADOclasses.h"
#include "DConfigGlobals.h" // For RunTimeError()
#ifdef ADO_SUPPORT // Possibly defined in DConfigGlobals.h
#include "adoce_i.c"

//////////////////////////////////////////////////////////////

long TComObject::m_RefCount = 0;

TComObject::TComObject()
{
	InterlockedIncrement(&m_RefCount);
	if (m_RefCount == 1) {
		HRESULT h = CoInitializeEx(NULL, COINIT_MULTITHREADED);
		if ((h != S_OK) && (h != S_FALSE)) {
			RunTimeError(_T("CoInitialize failed"));
		}
	}
}

TComObject::~TComObject()
{
	InterlockedDecrement(&m_RefCount);
	if (m_RefCount == 0) {
		CoUninitialize();
	}
}

CString TComObject::GetLastError()
{
	return m_LastError;
}

/////////////////////////////////////////////////////////

TADOField::TADOField()
{
	m_pField = NULL;
}

TADOField::~TADOField()
{
	if (m_pField) m_pField->Release();
}

bool TADOField::GetActualSize(long* size)
{
	HRESULT h = m_pField->get_ActualSize(size);
	if (FAILED(h)) {
		m_LastError.Format(_T("%d"), h);
		return false;
	}
	return true;
}

bool TADOField::GetDefinedSize(long* size)
{
	HRESULT h = m_pField->get_DefinedSize(size);
	if (FAILED(h)) {
		m_LastError.Format(_T("%d"), h);
		return false;
	}
	return true;
}

bool TADOField::GetName(CString* name)
{
	BSTR str;
	HRESULT h = m_pField->get_Name(&str);
	if (FAILED(h)) {
		m_LastError.Format(_T("%d"), h);
		return false;
	}
	*name = (LPCWSTR)str;
	SysFreeString(str);
	return true;
}

bool TADOField::GetType(DataTypeEnum* type)
{
	HRESULT h = m_pField->get_Type(type);
	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	return true;
}

bool TADOField::GetValue(VARIANT* value)
{
	HRESULT h = m_pField->get_Value(value);
	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	return true;
}

bool TADOField::SetValue(VARIANT value)
{
	HRESULT h = m_pField->put_Value(value);
	if (FAILED(h)) {
		m_LastError = _T("Failed to set field property");
		return false;
	}
	return true;
}

/////////////////////////////////////////////////////////

TADOFields::TADOFields()
{
	m_pFields = NULL;
}

TADOFields::~TADOFields()
{
	if (m_pFields) m_pFields->Release();
}

bool TADOFields::GetCount(long* count)
{
	HRESULT h = m_pFields->get_Count(count);
	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	return true;
}

bool TADOFields::GetItem(long index, TADOField* p)
{
	VARIANT varIndex;
	VariantInit(&varIndex);
	varIndex.vt = VT_I4;
	varIndex.lVal = index;

	HRESULT h = m_pFields->get_Item(varIndex, &p->m_pField);
	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	return true;
}

bool TADOFields::GetItem(LPCTSTR name, TADOField* p)
{
	VARIANT varIndex;
	VariantInit(&varIndex);
	varIndex.vt = VT_BSTR;
	varIndex.bstrVal = ::SysAllocString(name);

	HRESULT h = m_pFields->get_Item(varIndex, &p->m_pField);
	VariantClear(&varIndex);

	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	return true;
}

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

TADORecordset::TADORecordset()
{
	HRESULT h = CoCreateInstance(CLSID_ADOCERecordset, NULL, CLSCTX_INPROC_SERVER,
		IID_IADOCERecordset, (void**)&m_pRs);

	if (FAILED(h)) {
		RunTimeError(_T("Failed to create recordset COM interface"));
	}
}

TADORecordset::~TADORecordset()
{
	// Should close?
	if (m_pRs) m_pRs->Release();
}

bool TADORecordset::Open(LPCTSTR command, LPCTSTR con, long cursorType, long lockType,
						 long options)
{
	VARIANT vCommand, vCon;
	VariantInit(&vCommand);
	VariantInit(&vCon);

	vCommand.vt = VT_BSTR;
	vCommand.bstrVal = SysAllocString(command);

	vCon.vt = VT_BSTR;
	vCon.bstrVal = SysAllocString(con);

	HRESULT h = m_pRs->Open(vCommand, vCon, static_cast<CursorTypeEnum>(cursorType),
		static_cast<LockTypeEnum>(lockType), options);

	VariantClear(&vCommand);
	VariantClear(&vCon);

	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	return true;
}

bool TADORecordset::Close()
{
	HRESULT h = m_pRs->Close();
	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	return true;
	//if (m_pRs->State != adStateClosed) {
	//	m_pRs->Close();
	//}
}

bool TADORecordset::GetFields(TADOFields* fields)
{
	HRESULT h = m_pRs->get_Fields(&fields->m_pFields);
	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	return true;
}

bool TADORecordset::AddNew()
{
	VARIANT fieldList, values;
	fieldList.vt = VT_ERROR;
	values.vt = VT_ERROR;
	fieldList.scode = DISP_E_PARAMNOTFOUND;
	values.scode = DISP_E_PARAMNOTFOUND;

	HRESULT h = m_pRs->AddNew(fieldList, values);
	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	return true;
}

bool TADORecordset::Delete(long affect)
{
	affect = 1; // Must be 1 for CE?

	bool bof, eof;
	GetBOF(&bof);
	GetEOF(&eof);
	if (bof && eof) {
		m_LastError = _T("No records in recordset to delete");
		return false;
	}

	HRESULT h = m_pRs->Delete(affect);
	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	return true;
}

/*bool TADORecordset::Find(LPCTSTR criteria, long skipRecs,
						 SearchDirectionEnum srchDir, _variant_t start)
{
	if (FAILED(m_pRs->Find(criteria, skipRecs, srchDir, start))) {
		bResult = false;
		m_LastError.Empty();
	}
}*/

bool TADORecordset::GetRows()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::GetString(long stringFormat, long numRows, LPCTSTR colDelimeter,
							  LPCTSTR rowDelimeter, LPCTSTR nullExp, CString* result)
{
	//*result = (LPCTSTR)m_pRs->GetString(static_cast<StringFormatEnum>(stringFormat), numRows,
	//	colDelimeter, rowDelimeter, nullExp);
	return false;
}

bool TADORecordset::Move()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::MoveFirst()
{
	HRESULT h = m_pRs->MoveFirst();
	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	return true;
}

bool TADORecordset::MoveLast()
{
	HRESULT h = m_pRs->MoveLast();
	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	return true;
}

bool TADORecordset::MoveNext()
{
	HRESULT h = m_pRs->MoveNext();
	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	return true;
}

bool TADORecordset::MovePrevious()
{
	HRESULT h = m_pRs->MovePrevious();
	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	return true;
}

bool TADORecordset::Requery()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::Resync()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::Save()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::Supports()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::Update()
{
	VARIANT fields;
	VARIANT values;
	fields.vt = VT_ERROR;
	values.vt = VT_ERROR;
	fields.scode = DISP_E_PARAMNOTFOUND;
	values.scode = DISP_E_PARAMNOTFOUND;

	HRESULT h = m_pRs->Update(fields, values);
	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	return true;
}

bool TADORecordset::CancelUpdate()
{
	HRESULT h = m_pRs->CancelUpdate();
	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	return true;
}

bool TADORecordset::GetAbsolutePage()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::SetAbsolutePage()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::GetAbsolutePosition(long* pos)
{
	HRESULT h = m_pRs->get_AbsolutePosition(pos);
	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	return true;
}

bool TADORecordset::SetAbsolutePosition(long pos)
{
	HRESULT h = m_pRs->put_AbsolutePosition(pos);
	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	return true;
}

bool TADORecordset::GetBOF(bool* bBOF)
{
	VARIANT_BOOL var;
	HRESULT h = m_pRs->get_BOF(&var);
	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	*bBOF = var == VARIANT_TRUE;
	return true;
}

bool TADORecordset::GetEOF(bool* bEOF)
{
	VARIANT_BOOL var;
	HRESULT h = m_pRs->get_EOF(&var);
	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	*bEOF = var == VARIANT_TRUE;
	return true;
}

#ifndef UNDER_CE
bool TADORecordset::GetCursorLocation()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::SetCursorLocation()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}
#endif

bool TADORecordset::GetCursorType()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::SetCursorType()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::GetEditMode()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::SetEditMode()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::GetLockType()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::SetLockType()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::GetRecordCount(long* size)
{
	HRESULT h = m_pRs->get_RecordCount(size);
	if (FAILED(h)) {
		m_LastError.Empty();
		return false;
	}
	return true;
}

bool TADORecordset::GetSort()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::SetSort()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::GetSource()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::SetSource()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::GetState()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

bool TADORecordset::GetStatus()
{
	m_LastError = _T("Unimplemented function or property");
	return false;
}

////////////////////////////////
#endif // ADO_SUPPORT
