/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

#include "stdafx.h"
#include "Base64.h"

static unsigned char encodeTable[65] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static unsigned char decodeTable[] = {
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0x3E,0xFF,0xFF,0xFF,0x3F,
	0x34,0x35,0x36,0x37,0x38,0x39,0x3A,0x3B,0x3C,0x3D,0xFF,0xFF,0xFF,0x00,0xFF,0xFF,
	0xFF,0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0x0C,0x0D,0x0E,
	0x0F,0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0xFF,0xFF,0xFF,0xFF,0xFF,
	0xFF,0x1A,0x1B,0x1C,0x1D,0x1E,0x1F,0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,
	0x29,0x2A,0x2B,0x2C,0x2D,0x2E,0x2F,0x30,0x31,0x32,0x33,0xFF,0xFF,0xFF,0xFF,0xFF
};

bool DecodeBase64(TString text, VALUE& vChunkOut)
{
	LPCTSTR pText = text;
	int textLength = text.GetLength();
	int outBufLength = 0;
	int outBufMaxLength = ((textLength + 3) / 4) * 3;
	unsigned char* pOutBuf = new unsigned char[outBufMaxLength];
	unsigned char* pOutBufHead = pOutBuf;

	unsigned char ch, dch;
	unsigned int bits = 0;
	int padCount = 0;
	int bitCount = 0;
	for (int idx = 0; idx < textLength; ++idx) {
		ch = (*pText++) & 0x7f;
		dch = decodeTable[ch];
		if (dch != 0xFF) {
			if (ch == '=') {
				++padCount;
			}
			bits = (bits << 6) | dch;
			bitCount += 6;
			if (bitCount >= 8) {
				bitCount -= 8;
				*pOutBuf++ = (bits >> bitCount) & 0xff;
				bits &= ((1 << bitCount) - 1);
				outBufLength++;
			}
		}
	}

	outBufLength -= padCount;
	vChunkOut = AllocChunk(outBufLength, pOutBufHead);
	delete[] pOutBufHead;
	return true;
}

bool EncodeBase64(const unsigned char *data, int length, TString& textOut)
{
	TCHAR* pOutBuf = textOut.GetBuffer(length * 2 + 4);
	TCHAR* pOutBufHead = pOutBuf;

	int columns = 0;	// Put carraige return every 76 columns
	unsigned long bits;
	int idx = 0;
	while (length - idx >= 3) {
		bits = (data[idx] << 16) | (data[idx + 1] << 8) | data[idx + 2];
		idx += 3;
		*pOutBuf++ = encodeTable[(bits >> 18) & 0x3f];
		*pOutBuf++ = encodeTable[(bits >> 12) & 0x3f];
		*pOutBuf++ = encodeTable[(bits >> 6) & 0x3f];
		*pOutBuf++ = encodeTable[bits & 0x3f];

		if ((columns += 4) >= 75) {
			columns = 0;
			#ifndef __UNIX__
				*pOutBuf++ = '\r';
			#endif
			#ifndef __MAC__
				*pOutBuf++ = '\n';
			#endif
		}
	}

	// Finish up the last few characters
	if (length - idx == 2) {
		bits = (data[idx] << 16) | (data[idx + 1] << 8);
		*pOutBuf++ = encodeTable[(bits >> 18) & 0x3f];
		*pOutBuf++ = encodeTable[(bits >> 12) & 0x3f];
		*pOutBuf++ = encodeTable[(bits >> 6) & 0x3f];
		*pOutBuf++ = '=';
	}
	else if (length - idx == 1) {
		bits = (data[idx] << 16);
		*pOutBuf++ = encodeTable[(bits >> 18) & 0x3f];
		*pOutBuf++ = encodeTable[(bits >> 12) & 0x3f];
		*pOutBuf++ = '=';
		*pOutBuf++ = '=';
	}

	textOut.ReleaseBuffer(pOutBuf - pOutBufHead);
	return true;
}

/////////////////////////////////////////////////////////////

bool DecodeHex(TString text, VALUE& vChunkOut)
{
	// x = [:{encoding:"hex"}, "a8f1234":]

	int textLength = text.GetLength();
	LPCTSTR pTextHead = text;
	LPCTSTR pText = pTextHead + textLength - 1; // Walk backwards

	int OutBufLength = textLength / 2 + 1;		// Possible over estimate
	unsigned char *pOutBufHead = new unsigned char[OutBufLength];
	unsigned char *pOutBuf = pOutBufHead + OutBufLength;		// Walk backwards

	TCHAR ch;
	unsigned char bits = 0, val;
	OutBufLength = 0;
	while (pText >= pTextHead) {
		val = 0;
		ch = *pText--;
		if (ch <= ' ') continue;
		else if ((ch >= '0') && (ch <= '9')) {
			val = ch - '0';
		}
		else if ((ch >= 'a') && (ch <= 'f')) {
			val = ch - 'a' + 10;
		}
		else if ((ch >= 'A') && (ch <= 'F')) {
			val = ch - 'A' + 10;
		}
		else {
			vChunkOut = VALUE();
			delete[] pOutBufHead;
			return false;
		}

		if (bits == 0) {
			--pOutBuf;
			++OutBufLength;
			*pOutBuf = val;
			bits = 4;
		}
		else {
			*pOutBuf = *pOutBuf | (val << 4);
			bits = 0;
		}
	}

	vChunkOut = AllocChunk(OutBufLength, pOutBuf);
	delete[] pOutBufHead;
	return true;
}

static TCHAR tohexTable[16] = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};

bool EncodeHex(const unsigned char *data, int length, TString& textOut)
{
	textOut.Empty();
	int column = 0;
	for (int idx = 0; idx < length; ++idx) {
		textOut += tohexTable[(data[idx] & 0xF0) >> 4];
		textOut += tohexTable[data[idx] & 0xF];
		column += 2;
		if (column >= 76) {
			// Carriage return every 76 chars
			column = 0;
			#ifndef __UNIX__
				textOut += '\r';
			#endif
			#ifndef __MAC__
				textOut += '\n';
			#endif
		}
	}
	return true;
}
