/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

#include "stdafx.h"
#include "DConfigGlobals.h"
#include "CDebug.h"
#include <stdarg.h>	// For va_start, etc.

///////////////////////////////////
#if defined(USING_MFC) && defined(_DEBUG)
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif
///////////////////////////////////

//===============================================================

void SyntaxError(LPCTSTR parmString, ...)
{
	TCHAR buf[512];
	va_list rest;
	va_start(rest, parmString);
	_vstprintf(buf, parmString, rest);
	va_end(rest);
	RAISE_SYNTAX(buf, _T("__core__"), 0);
}

void RunTimeError(LPCTSTR parmString, ...)
{
	TCHAR buf[512];
	va_list rest;
	va_start(rest, parmString);
	_vstprintf(buf, parmString, rest);
	va_end(rest);
	RAISE_RUNTIME(buf, _T("__core__"), 0);
}

void InternalError(LPCTSTR parmString, ...)
{
	ASSERT(false);	// Stop debugger before the exception is thrown

	TCHAR buf[512];
	va_list rest;
	va_start(rest, parmString);
	_vstprintf(buf, parmString, rest);
	va_end(rest);
	RAISE_INTERPRETER(buf, _T("__core__"), 0);
}

//===============================================================

EException::EException(LPCTSTR _module, int _line, DWORD exceptCode, LPCTSTR msg)
{
	exceptionCode = exceptCode;
	line = _line;
	SetStrings(_module, msg);
	//module = NULL;
	//message = NULL;
}

EException::EException(const EException& copy)
{
	exceptionCode = copy.exceptionCode;
	line = copy.line;
	SetStrings(copy.module, copy.message);
	//module = NULL;
	//message = NULL;
}

EException& EException::operator = (const EException& copy)
{
	if (this != &copy) {
		if (message) delete[] message;
		if (module) delete[] module;

		exceptionCode = copy.exceptionCode;
		line = copy.line;
		SetStrings(copy.module, copy.message);
		//module = NULL;
		//message = NULL;
	}
	return *this;
}

EException::~EException()
{
	if (message) delete[] message;
	if (module) delete[] module;
}

void EException::SetStrings(LPCTSTR mod, LPCTSTR msg)
{
	if (msg == NULL) message = NULL;
	else {
		int length = _tcslen(msg);
		message = new TCHAR[length + 1];
		_tcscpy(message, msg);
	}

	if (mod == NULL) module = NULL;
	else {
		int length = _tcslen(mod);
		module = new TCHAR[length + 1];
		_tcscpy(module, mod);
	}
}

//===============================================================
