/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

#include "stdafx.h"
#include "DConfigGlobals.h"
#include "CExceptionMod.h"
#include "Interpreter.h"

///////////////////////////////////
#if defined(USING_MFC) && defined(_DEBUG)
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif
///////////////////////////////////

// These statics are for use as the base of derived exceptions
static VALUE vEExceptionClass = AllocClass(new CEException());
static VALUE vESocketClass = AllocClass(new CESocket());

//===============================================================

VALUE InitExceptionModule()
{
	VALUE symtab = AllocFrame(7);
	symtab.Frame()->Add(_T("__module__"), (LPCTSTR)EXCEPTION_MODULE_NAME);
	symtab.Frame()->Add(_T("EException"), vEExceptionClass);
	symtab.Frame()->Add(_T("ESyntax"), AllocClass(new CESyntax()));
	symtab.Frame()->Add(_T("ERuntime"), AllocClass(new CERuntime()));
	symtab.Frame()->Add(_T("ESocket"), AllocClass(new CESocket()));
	symtab.Frame()->Add(_T("ETimeout"), AllocClass(new CETimeout()));
	return symtab;
}

//===============================================================

CEException::CEException():CClassImpl(VALUE(), _T("EException"))
{
	m_MemberSymTab = AllocFrame(19);
	// Anything added here is static (becomes shared by all instances of the class)
	m_MemberSymTab.Set(_T("__type__"), _T("eexception"));
	m_MemberSymTab.Set(_T("__init__"), AllocFunction(__init__, (LPCTSTR)EXCEPTION_MODULE_NAME));
}

CEException::~CEException(){}

VALUE CEException::__init__(VALUE* self, int argCount, VALUE args[], CInterpreter* pInterp)
{
	if (argCount != 2) {
		SyntaxError(_T("eexception.__init__ requires two args"));
	}
	else {
		self->Set(_T("msg"), args[0]);
		self->Set(_T("other"), args[1]);
		self->Set(_T("module"), pInterp->CurrentModuleName());
		self->Set(_T("line"), pInterp->CurrentLine());
	}

	return VALUE();
}

CESyntax::CESyntax():CClassImpl(vEExceptionClass, _T("ESyntax"))
{
	m_MemberSymTab = AllocFrame(19);
	// Anything added here is static (becomes shared by all instances of the class)
	m_MemberSymTab.Set(_T("__type__"), _T("esyntax"));
	m_MemberSymTab.Set(_T("__init__"), AllocFunction(__init__, (LPCTSTR)EXCEPTION_MODULE_NAME));
}

CESyntax::~CESyntax(){}

VALUE CESyntax::__init__(VALUE* self, int argCount, VALUE args[], CInterpreter* pInterp)
{
	if (argCount != 2) {
		SyntaxError(_T("esyntax.__init__ requires two args"));
	}
	else {
		self->Set(_T("msg"), args[0]);
		self->Set(_T("other"), args[1]);
		self->Set(_T("module"), pInterp->CurrentModuleName());
		self->Set(_T("line"), pInterp->CurrentLine());
	}
	return VALUE();
}

CERuntime::CERuntime():CClassImpl(vEExceptionClass, _T("ERuntime"))
{
	m_MemberSymTab = AllocFrame(19);
	// Anything added here is static (becomes shared by all instances of the class)
	m_MemberSymTab.Set(_T("__type__"), _T("eruntime"));
	m_MemberSymTab.Set(_T("__init__"), AllocFunction(__init__, (LPCTSTR)EXCEPTION_MODULE_NAME));
}

CERuntime::~CERuntime(){}

VALUE CERuntime::__init__(VALUE* self, int argCount, VALUE args[], CInterpreter* pInterp)
{
	if (argCount != 2) {
		SyntaxError(_T("eruntime.__init__ requires two args"));
	}
	else {
		self->Set(_T("msg"), args[0]);
		self->Set(_T("other"), args[1]);
		self->Set(_T("module"), pInterp->CurrentModuleName());
		self->Set(_T("line"), pInterp->CurrentLine());
	}
	return VALUE();
}

CEInternal::CEInternal():CClassImpl(vEExceptionClass, _T("EInternal"))
{
	m_MemberSymTab = AllocFrame(19);
	// Anything added here is static (becomes shared by all instances of the class)
	m_MemberSymTab.Set(_T("__type__"), _T("einternal"));
	m_MemberSymTab.Set(_T("__init__"), AllocFunction(__init__, (LPCTSTR)EXCEPTION_MODULE_NAME));
}

CEInternal::~CEInternal(){}

VALUE CEInternal::__init__(VALUE* self, int argCount, VALUE args[], CInterpreter* pInterp)
{
	if (argCount != 2) {
		SyntaxError(_T("einternal.__init__ requires two args"));
	}
	else {
		self->Set(_T("msg"), args[0]);
		self->Set(_T("other"), args[1]);
		self->Set(_T("module"), pInterp->CurrentModuleName());
		self->Set(_T("line"), pInterp->CurrentLine());
	}
	return VALUE();
}

CESocket::CESocket():CClassImpl(vEExceptionClass, _T("ESocket"))
{
	m_MemberSymTab = AllocFrame(19);
	// Anything added here is static (becomes shared by all instances of the class)
	m_MemberSymTab.Set(_T("__type__"), _T("esocket"));
	m_MemberSymTab.Set(_T("__init__"), AllocFunction(__init__, (LPCTSTR)EXCEPTION_MODULE_NAME));
}

CESocket::~CESocket(){}

VALUE CESocket::__init__(VALUE* self, int argCount, VALUE args[], CInterpreter* pInterp)
{
	if (argCount != 2) {
		SyntaxError(_T("esocket.__init__ requires two args"));
	}
	else {
		self->Set(_T("msg"), args[0]);
		self->Set(_T("other"), args[1]);
		self->Set(_T("module"), pInterp->CurrentModuleName());
		self->Set(_T("line"), pInterp->CurrentLine());
	}
	return VALUE();
}

CETimeout::CETimeout():CClassImpl(vESocketClass, _T("ETimeout"))
{
	m_MemberSymTab = AllocFrame(19);
	// Anything added here is static (becomes shared by all instances of the class)
	m_MemberSymTab.Set(_T("__type__"), _T("etimeout"));
	m_MemberSymTab.Set(_T("__init__"), AllocFunction(__init__, (LPCTSTR)EXCEPTION_MODULE_NAME));
}

CETimeout::~CETimeout(){}

VALUE CETimeout::__init__(VALUE* self, int argCount, VALUE args[], CInterpreter* pInterp)
{
	if (argCount != 2) {
		SyntaxError(_T("etimeout.__init__ requires two args"));
	}
	else {
		self->Set(_T("msg"), args[0]);
		self->Set(_T("other"), args[1]);
		self->Set(_T("module"), pInterp->CurrentModuleName());
		self->Set(_T("line"), pInterp->CurrentLine());
	}
	return VALUE();
}
