/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

#include "stdafx.h"
#include "DConfigGlobals.h"
#include "CTokenTypes.h"

///////////////////////////////////
#if defined(USING_MFC) && defined(_DEBUG)
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif
///////////////////////////////////

struct TokStringPair {
	LPCTSTR String;
	TokenTypes TokenType;
};

/*
 *	KeywordStrings holds all predefined keywords that can be encountered
 *	in a program such as flow control and math functions.
 *	This is how we convert from string to token type.  Ordering this
 *	list based on usage can help speed things up.
 */
static TokStringPair KeywordStrings[] = {
	{_T("if"), tokIf}, {_T("then"), tokThen}, {_T("else"), tokElse},
	{_T("elif"), tokElseIf}, {_T("elseif"), tokElseIf},	{_T("endif"), tokEndIf},
	{_T("for"), tokFor}, {_T("step"), tokStep}, {_T("endfor"), tokEndFor}, {_T("next"), tokEndFor},
	{_T("foreach"), tokForEach}, {_T("endforeach"), tokEndForEach},
	{_T("while"), tokWhile}, {_T("endwhile"), tokEndWhile},
	{_T("class"), tokClass}, {_T("endclass"), tokEndClass},
	{_T("inherited"), tokInherited},
	{_T("and"), tokAnd}, {_T("or"), tokOr}, {_T("not"), tokNot},
	{_T("print"), tokPrint}, {_T("cr"), tokCR},
	{_T("in"), tokIn}, {_T("return"), tokReturn},
	{_T("with"), tokWith}, {_T("endwith"), tokEndWith}, {_T("map"), tokMap},
	{_T("throw"), tokThrow}, {_T("try"), tokTry}, {_T("endtry"), tokEndTry},
	{_T("catch"), tokCatch}, {_T("finally"), tokFinally},
	{_T("func"), tokFunc}, {_T("endfunc"), tokEndFunc},
	//{_T("function"), tokFunc}, {_T("endfunction"), tokEndFunc},
	{_T("break"), tokBreak}, {_T("continue"), tokContinue}, {_T("import"), tokImport},
	{_T("to"), tokTo}, //{_T("from"), tokFrom},
	{_T("type"), tokType}, {_T("run"), tokRun},
	{_T("left$"), tokLeft_}, {_T("right$"), tokRight_},
	{_T("eval"), tokEvaluate}, {_T("defined"), tokDefined},

	{_T("true"), tokTrue}, {_T("false"), tokFalse}, {_T("xor"), tokXOR}, {_T("nil"), tokNil},

	{_T("sectitem"), tokSItem}, {_T("startsect"), tokStartSect}, {_T("endsect"), tokEndSect},

	{_T("len"), tokLen}, {_T("tab"), tokTab},
	{_T("str$"), tokStr_}, {_T("chr$"), tokChr_},
	//{_T("date$"), tokDate_}, {_T("longdate$"), tokLongDate_}, {_T("time$"), tokTime_},
	//{_T("pos"), tokPos}, {_T("cap$"), tokCap_}, {_T("upper$"), tokUpper_}, {_T("lower$"), tokLower_},
	{_T("prop"), tokProp},
	{_T("as"), tokAs}, {_T("debug"), tokDebugOut}, {_T("isa"), tokIsa}, {_T("div"), tokIntDivide}, // added Guilect
	{_T("do"), tokDo}, {_T("local"), tokLocal}, {_T("interpret"), tokInterpret}, {_T("pow"), tokPowerOp} 
};

/*
 *	Used primarily for unparsing
 */
static TokStringPair OperatorStrings[] = {
	{_T(","), tokComma}, {_T("."), tokDot},
	{_T("+"), tokPlus}, {_T("-"), tokMinus}, {_T("*"), tokTimes}, {_T("/"), tokDivide},
	{_T(":"), tokColon}, {_T("<>"), tokNotEquals}, {_T("><"), tokNotEqualsCase},
	{_T("["), tokOpenBracket}, {_T("]"), tokCloseBracket},
	{_T("{"), tokOpenBrace}, {_T("}"), tokCloseBrace},
	{_T("("), tokOpenParen}, {_T(")"), tokCloseParen},
	{_T(";"), tokSemicolon}, {_T("'"), tokTick}, {_T("?"), tokQuestion},
	{_T("\\\\"), tokPersistComment}, {_T("\\"), tokBackslash},
	{_T(">>"), tokLeftShift}, {_T("<<"), tokRightShift},
	{_T(">"), tokGreater}, {_T("<"), tokLess},
	{_T(">="), tokGreaterEqual}, {_T("<="), tokLessEqual},
	{_T("^"), tokPower},  {_T("#"), tokPound}, {_T("="), tokAssignment},
	{_T("!="), tokNotEquals}, {_T("&"), tokBitAnd}, {_T("|"), tokBitOr}, {_T("!"), tokBitNot},
	{_T("_"), tokUnderscore}, {_T("~"), tokTilde}, {_T("@"), tokAt},
	{_T("%"), tokPercent}, {_T("->"), tokRange},
	{_T("[:"), tokStartChunk}, {_T(":]"), tokEndChunk},  {_T("=="), tokBooleanEquals}	//added tokBooleanEquals 2/10/03  Guilect
};

/*
 *	OtherStrings is only used for error testing.  Provides a means
 *	to print out "non-printable" token types, etc.
 */
static TokStringPair OtherStrings[] = {
	{_T("EOF"), tokEOF}, {_T("identifier"), tokIdent}, {_T("float"), tokFloat},
	{_T("integer"), tokInteger},
	{_T("quoted-string"), tokQuotedString},	{_T("undefined"), tokUndefined},
	{_T("EOL"), tokEOL},
};


inline int NumberKeywordStringEntries()
{
	return sizeof(KeywordStrings) / sizeof(TokStringPair);
}

inline int NumberOperatorStringEntries()
{
	return sizeof(OperatorStrings) / sizeof(TokStringPair);
}

inline int NumberOtherStringEntries()
{
	return sizeof(OtherStrings) / sizeof(TokStringPair);
}

//=================================================================

TokenTypes StringToTokenType(TString str)
{
	int length = NumberKeywordStringEntries();
	for (int idx = 0; idx < length; idx++) {
		if (str.CompareNoCase(KeywordStrings[idx].String) == 0) {
			return KeywordStrings[idx].TokenType;
		}
	}
	return tokIdent;	// If it's not a keyword, then it's just an ident
}

TString TokenTypeToKeywordString(TokenTypes tok)
{
	int idx;
	int length = NumberKeywordStringEntries();
	for (idx = 0; idx < length; idx++) {
		if (tok == KeywordStrings[idx].TokenType) {
			return KeywordStrings[idx].String;
		}
	}
	return _T("");
}

TString TokenTypeToString(TokenTypes tok)
{
	int idx;
	int length = NumberKeywordStringEntries();

	for (idx = 0; idx < length; idx++) {
		if (tok == KeywordStrings[idx].TokenType) {
			return KeywordStrings[idx].String;
		}
	}

	length = NumberOperatorStringEntries();
	for (idx = 0; idx < length; idx++) {
		if (tok == OperatorStrings[idx].TokenType) {
			return OperatorStrings[idx].String;
		}
	}

	length = NumberOtherStringEntries();
	for (idx = 0; idx < length; idx++) {
		if (tok == OtherStrings[idx].TokenType) {
			return OtherStrings[idx].String;
		}
	}
	return TString(_T(" "));
}
