/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

#ifndef _CTOKENTYPES__H__
#define _CTOKENTYPES__H__

#include "DConfigGlobals.h"

////////////////////////////////////////////////////////////////
// Changing the order of any of the enumerations will cause any currently
// compiled (and serialized data with functions) to be invalid!!!
// Adding new tokens to the end should be alright since that
// won't change the values of the previous tokens.  Of course,
// then an older version of the interpreter won't recognize the
// new tokens.  If you change anything, be sure to make exactly
// the same changes in the ByteCodeToType array below.
////////////////////////////////////////////////////////////////
enum TokenTypes {
	tokUndefined, tokIdent, tokFloat, tokInteger, tokQuotedString, tokError,
	tokEOF, tokEOL,	tokNil,

	tokStr_, tokLen, tokDate_, tokLongDate_, tokTime_, tokPos, tokCap_,
	tokUpper_, tokLower_, tokRight_, tokChr_,

	tokIf, tokThen, tokElse, tokElseIf, tokEndIf, tokWhile, tokEndWhile, tokCase, tokEndCase,
	tokClass, tokEndClass, tokConst, tokIn, tokReturn,
	tokCatch, tokFinally, tokFor, tokEndFor, tokForEach,  tokEndForEach,
	tokFunc, tokEndFunc, tokBreak, tokContinue, tokImport, tokTo,
	tokNewPage, tokMessage, tokTable, tokEndTable,
	tokRead, tokFrom, tokPrint, tokCR, tokRun,
	tokEvaluate,

	tokTick, tokOpenParen, tokCloseParen, tokDot, tokQuestion, tokUnderscore,
	tokOpenBracket, tokCloseBracket, tokOpenBrace, tokCloseBrace, tokAssignment,
	tokDoubleQuote, tokSingleQuote, tokPlusEqual, tokMinusEqual,
	tokTimesEqual, tokDivideEqual, tokPound, tokComma, tokColon, tokSemicolon,
	tokTilde, tokAt,

	tokTrue, tokFalse,
	tokNot, tokPower,
	tokBitAnd, tokBitOr, tokBitNot, tokXOR, tokLeftShift, tokRightShift,
	tokOr, tokPlus, tokMinus, tokAnd, tokTimes, tokDivide,
	tokBooleanEquals, tokNotEquals,
	tokGreater, tokLess, tokGreaterEqual, tokLessEqual,

	tokStartSection, tokEndSection, tokSItem, tokWith, tokEndWith,
	tokTab, tokType, tokLeft_, tokMap, tokSort,
	tokPersistComment, tokBackslash,
	tokProp, tokAs, tokThrow, tokTry, tokEndTry,
	tokThis, tokInherited, tokStartSect, tokEndSect,
	tokDebugOut, tokIsa, tokDefined, tokDo, tokLocal, tokPercent,
	tokNotEqualsCase, tokRange, tokStartChunk, tokEndChunk, tokInterpret, tokStep,
	tokIntDivide, tokPowerOp	// added Guilect
};


TokenTypes StringToTokenType(TString str);
TString TokenTypeToString(TokenTypes tok);

/*
 *	These MUST be in the same order as the TokenTypes enumeration decl.
 */
static TokenTypes ByteCodeToType[256] = {
	tokUndefined, tokIdent, tokFloat, tokInteger, tokQuotedString, tokError,
	tokEOF, tokEOL,	tokNil,

	tokStr_, tokLen, tokDate_, tokLongDate_, tokTime_, tokPos, tokCap_,
	tokUpper_, tokLower_, tokRight_, tokChr_,

	tokIf, tokThen, tokElse, tokElseIf, tokEndIf, tokWhile, tokEndWhile, tokCase, tokEndCase,
	tokClass, tokEndClass, tokConst, tokIn, tokReturn,
	tokCatch, tokFinally, tokFor, tokEndFor, tokForEach,  tokEndForEach,
	tokFunc, tokEndFunc, tokBreak, tokContinue, tokImport, tokTo,
	tokNewPage, tokMessage, tokTable, tokEndTable,
	tokRead, tokFrom, tokPrint, tokCR, tokRun,
	tokEvaluate,

	tokTick, tokOpenParen, tokCloseParen, tokDot, tokQuestion, tokUnderscore,
	tokOpenBracket, tokCloseBracket, tokOpenBrace, tokCloseBrace, tokAssignment,
	tokDoubleQuote, tokSingleQuote, tokPlusEqual, tokMinusEqual,
	tokTimesEqual, tokDivideEqual, tokPound, tokComma, tokColon, tokSemicolon,
	tokTilde, tokAt,

	tokTrue, tokFalse,
	tokNot, tokPower,
	tokBitAnd, tokBitOr, tokBitNot, tokXOR, tokLeftShift, tokRightShift,
	tokOr, tokPlus, tokMinus, tokAnd, tokTimes, tokDivide,
	tokBooleanEquals, tokNotEquals,
	tokGreater, tokLess, tokGreaterEqual, tokLessEqual,

	tokStartSection, tokEndSection, tokSItem, tokWith, tokEndWith,
	tokTab, tokType, tokLeft_, tokMap, tokSort,
	tokPersistComment, tokBackslash,
	tokProp, tokAs, tokThrow, tokTry, tokEndTry,
	tokThis, tokInherited, tokStartSect, tokEndSect,
	tokDebugOut, tokIsa, tokDefined, tokDo, tokLocal, tokPercent,
	tokNotEqualsCase, tokRange, tokStartChunk, tokEndChunk, tokInterpret, tokStep,
	tokIntDivide, tokPowerOp	// added Guilect
};

//===============================================================
#endif	// _CTOKENTYPES__H__
