/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

#include "stdafx.h"
#include "DConfigGlobals.h"

#ifdef MATH_MODULE_SUPPORT

#include "MathModule.h"
#include "Interpreter.h"
#include <math.h>

///////////////////////////////////
#if defined(_DEBUG) && (defined(__WINCE__) || defined(_WINNT__))
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif
///////////////////////////////////

#define XERR(x) pInterp->RuntimeError(_T(x))
#define ERR(x) {XERR(x);return AllocNil();}

#define REQUIRE_ARGS(x) if (argCount!=x) ERR("Wrong number of arguments to function.")
#define REQUIRE_STRING(x) if (args[x].Type() != stString) ERR("Expected string in argument(s) to function.")
#define REQUIRE_FLOAT(x) if ((args[x].Type() != stFloat)&&(args[x].Type() != stInteger)) ERR("Expected number in argument(s) to function.")
#define REQUIRE_INTEGER(x) if ((args[x].Type() != stInteger)) ERR("Expected integer in arguments(s) to function.")
#define REQUIRE_ARRAY(x) if ((args[x].Type() != stArray)) ERR("Expected array in arguments(s) to function.")

static LPCTSTR rof = _T("Expecting one floating point argument to function");
#define REQUIRE_ONE_FLOAT \
	if ((argCount != 1) || ((args[0].Type() != stFloat) && (args[0].Type() != stInteger))) \
		{pInterp->RuntimeError(rof); return AllocNil();}

CORE_FUNC(math_sin)
{
	REQUIRE_ONE_FLOAT;
	return (float)sin(args[0].Float());
}

CORE_FUNC(math_cos)
{
	REQUIRE_ONE_FLOAT;
	return (float)cos(args[0].Float());
}

CORE_FUNC(math_tan)
{
	REQUIRE_ONE_FLOAT;
	return (float)tan(args[0].Float());
}

CORE_FUNC(math_asin)
{
	REQUIRE_ONE_FLOAT;
	return (float)asin(args[0].Float());
}

CORE_FUNC(math_acos)
{
	REQUIRE_ONE_FLOAT;
	return (float)acos(args[0].Float());
}

CORE_FUNC(math_sinh)
{
	REQUIRE_ONE_FLOAT;
	return (float)sinh(args[0].Float());
}

CORE_FUNC(math_cosh)
{
	REQUIRE_ONE_FLOAT;
	return (float)cosh(args[0].Float());
}

CORE_FUNC(math_tanh)
{
	REQUIRE_ONE_FLOAT;
	return (float)tanh(args[0].Float());
}


CORE_FUNC(math_atan)
{
	REQUIRE_ONE_FLOAT;
	return (float)atan(args[0].Float());
}

static int factorial(int x)
{
	if (x <= 1) return 1;
	else return x * factorial(x - 1);
}

CORE_FUNC(math_factorial)
{
	REQUIRE_ONE_FLOAT;
	return factorial(args[0].Integer());
}

CORE_FUNC(math_rand)
{
	static bool bSeeded = false;
	if (!bSeeded) {
		bSeeded = true;
		srand(GetTickCount());
	}
	return (float)rand() / (float)(RAND_MAX + 1);	// Return 0.0 <= x < 1.0
}

CORE_FUNC(math_round)
{
	static bool bRoundUp = false;	// Alternate if .50000

	if ((argCount < 1) || (argCount > 2) || ((args[0].Type() != stFloat) && (args[0].Type() != stInteger))) {
		ERR("Invalid args to math.round(float, <option>)");
	}

	int option;
	double i, r;
	r = modf(args[0].Float(), &i);
	if (r == 0.0f) return args[0];	// Already rounded!

	// Options:
	//  RND_NORMAL		0
	//	RND_TOZERO		1
	//	RND_FROMZERO	2
	//	RND_TOGGLE		3
	if (argCount < 2) option = 0;
	else option = args[1].Integer();

	if (option == 1) {
		if (i >= 0.0f) return (float)i;
		else return (float)i;
	}
	else if (option == 2) {
		if (i >= 0.0f) return (float)i + 1.0f;
		else return (float)i - 1.0f;
	}
	else if (option == 3) {
		if (r == 0.5f) {
			bRoundUp = !bRoundUp;
			return (float)(bRoundUp ? i + 1.0f : i);
		}
		else return (float)(r < 0.5f ? i : i + 1.0f);
	}
	else {
		return (float)(r < 0.5f ? i : i + 1.0f);
	}
}

CORE_FUNC(math_exp)
{
	REQUIRE_ONE_FLOAT;
	return (float)exp(args[0].Float());
}

CORE_FUNC(math_log)
{
	REQUIRE_ONE_FLOAT;
	return (float)log10(args[0].Float());
}

CORE_FUNC(math_ln)
{
	REQUIRE_ONE_FLOAT;
	return (float)log(args[0].Float());
}

CORE_FUNC(math_sqrt)
{
	REQUIRE_ONE_FLOAT;
	return (float)sqrt(args[0].Float());
}

CORE_FUNC(math_div)
{
	// Pure integer division
	REQUIRE_ARGS(2);
	return args[0].Integer() / args[1].Integer();
}

CORE_FUNC(math_pow)
{
	REQUIRE_ARGS(2);
	REQUIRE_FLOAT(0);
	REQUIRE_FLOAT(1);
	return (float)pow(args[0].Float(), args[1].Float());
}

CORE_FUNC(math_and)
{
	REQUIRE_ARGS(2);
	REQUIRE_INTEGER(0);
	REQUIRE_INTEGER(1);
	return (int)((DWORD)(args[0].Integer()) & (DWORD)(args[1].Integer()));
}

CORE_FUNC(math_or)
{
	REQUIRE_ARGS(2);
	REQUIRE_INTEGER(0);
	REQUIRE_INTEGER(1);
	return (int)(((DWORD)args[0].Integer()) | ((DWORD)args[1].Integer()));
}

CORE_FUNC(math_not)
{
	REQUIRE_ARGS(1);
	REQUIRE_INTEGER(0);
	return (int)(~(DWORD)args[0].Integer());
}

CORE_FUNC(math_xor)
{
	REQUIRE_ARGS(2);
	REQUIRE_INTEGER(0);
	REQUIRE_INTEGER(1);
	return (int)(((DWORD)args[0].Integer()) ^ ((DWORD)args[1].Integer()));
}

CORE_FUNC(math_shl)
{
	REQUIRE_ARGS(2);
	REQUIRE_INTEGER(0);
	REQUIRE_INTEGER(1);
	return (int)(((DWORD)args[0].Integer()) << ((DWORD)args[1].Integer()));
}

CORE_FUNC(math_shr)
{
	REQUIRE_ARGS(2);
	REQUIRE_INTEGER(0);
	REQUIRE_INTEGER(1);
	return (int)(((DWORD)args[0].Integer()) >> ((DWORD)args[1].Integer()));
}

CORE_FUNC(math_gcd)
{
	if ((argCount != 2) || (args[0].Type() != stInteger) || (args[1].Type() != stInteger)) {
		ERR("math.gcd() requires two integer arguments");
	}

	int a = args[0].Integer();
	int b = args[1].Integer();
	if (a < 0) a = -a;
	if (b < 0) b = -b;

	if ((a == 0) && (b == 0)) return 0;

	int ax;
	while (b != 0) {
		ax = b;
		b = a % b;
		a = ax;
	}
	return a;
}

VALUE InitMathModule()
{
	VALUE module = AllocFrame();
	module.Set(_T("__module__"), MATH_MODULE_NAME);
	module.Set(_T("e"), (float)2.718282);
	module.Set(_T("pi"), (float)3.14159265358979);
	module.Set(_T("RND_NORMAL"), 0);
	module.Set(_T("RND_TOZERO"), 1);
	module.Set(_T("RND_FROMZERO"), 2);
	module.Set(_T("RND_TOGGLE"), 3);
	module.Set(_T("div"), AllocFunction(math_div, MATH_MODULE_NAME));
	module.Set(_T("sin"), AllocFunction(math_sin, MATH_MODULE_NAME));
	module.Set(_T("cos"), AllocFunction(math_cos, MATH_MODULE_NAME));
	module.Set(_T("tan"), AllocFunction(math_tan, MATH_MODULE_NAME));
	module.Set(_T("asin"), AllocFunction(math_asin, MATH_MODULE_NAME));
	module.Set(_T("acos"), AllocFunction(math_acos, MATH_MODULE_NAME));
	module.Set(_T("atan"), AllocFunction(math_atan, MATH_MODULE_NAME));
	module.Set(_T("sinh"), AllocFunction(math_sinh, MATH_MODULE_NAME));
	module.Set(_T("cosh"), AllocFunction(math_cosh, MATH_MODULE_NAME));
	module.Set(_T("tanh"), AllocFunction(math_tanh, MATH_MODULE_NAME));
	module.Set(_T("factorial"), AllocFunction(math_factorial, MATH_MODULE_NAME));
	module.Set(_T("rand"), AllocFunction(math_rand, MATH_MODULE_NAME));
	module.Set(_T("round"), AllocFunction(math_round, MATH_MODULE_NAME));
	module.Set(_T("exp"), AllocFunction(math_exp, MATH_MODULE_NAME));
	module.Set(_T("log"), AllocFunction(math_log, MATH_MODULE_NAME));
	module.Set(_T("ln"), AllocFunction(math_ln, MATH_MODULE_NAME));
	module.Set(_T("sqrt"), AllocFunction(math_sqrt, MATH_MODULE_NAME));
	module.Set(_T("pow"), AllocFunction(math_pow, MATH_MODULE_NAME));
	module.Set(_T("gcd"), AllocFunction(math_gcd, MATH_MODULE_NAME));
	module.Set(_T("band"), AllocFunction(math_and, MATH_MODULE_NAME));
	module.Set(_T("bor"), AllocFunction(math_or, MATH_MODULE_NAME));
	module.Set(_T("bnot"), AllocFunction(math_not, MATH_MODULE_NAME));
	module.Set(_T("bxor"), AllocFunction(math_xor, MATH_MODULE_NAME));
	module.Set(_T("bshl"), AllocFunction(math_shl, MATH_MODULE_NAME));
	module.Set(_T("bshr"), AllocFunction(math_shr, MATH_MODULE_NAME));
	return module;
}

//////////////////////////////////////////////////////////////////////
#endif // MATH_MODULE_SUPPORT