/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

#ifndef __TFILE_UNIX_H__
#define __TFILE_UNIX_H__

#include "DConfigGlobals.h"

#if defined(__WINNT__) || defined(__WINCE__)
#	ifndef USING_MFC	// MFC already includes <afxwin.h>
#		include <windows.h>
#	endif
#else
#	include <stdio.h>
#endif

class TFile {
public:
	TFile();
	~TFile();

	bool Open(LPCTSTR lpszFileName, int nOpenMods);
	void Close();

	static void Save(LPCTSTR fileName, TString data);
	static TString Load(LPCTSTR fileName);

	int GetFileType() const;
	unsigned long Seek(long offset, int from);
	unsigned long GetLength();
	unsigned long GetPosition();
	unsigned long Read(void* lpBuf, unsigned long nCount);
	unsigned long Write(const void* lpBuf, unsigned long nCount);
	static bool DeleteFile(LPCTSTR filename, bool bForce = false);
	static bool Exists(LPCTSTR filename);
	static bool NewerThan(LPCTSTR first, LPCTSTR second);
	static bool Move(LPCTSTR fileToMove, LPCTSTR newFileName);
	static bool Copy(LPCTSTR fileToCopy, LPCTSTR newFileName);
	static bool MakeDir(LPCTSTR dirName);
	static bool RemoveDir(LPCTSTR dirName);

	static int begin, current, end;	// Seek constants
	static int read, write, create, append, unicode, ascii, text, binary;	// Open modifiers

	// These funcs read and write unicode or ascii depending on m_FileType
	void Write(TString text);
	int Read(TString& result, int numChars);
	void WriteLine(TString text);
	int ReadLine(TString& result);
	bool SetEOLTerminator(TString term);
	TString GetEOLTerminator();
	bool IsEOF();

protected:
	int GuessTextFileType();	// Returns either unicode or ascii
	void WriteEOLTerminator();
	int ReadASCIILine(TString& result);
	int ReadUnicodeLine(TString& result);

	TCHAR* m_Filename;
	int m_FileType;				// Set to unicode, ascii, or binary
	TString m_EOLTerminator;	// Max length of two characters

#if defined(__WINNT__) || defined(__WINCE__)
	HANDLE m_File;
#else
	FILE* m_File;
#endif
};


#endif	// __TFILE_UNIX_H__
