/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

#ifndef _TSOCKET___H
#define _TSOCKET___H

#include "DConfigGlobals.h"
#include "CValue.h"

#if defined(__WINNT__) || defined(__WINCE__)
#	include <winsock.h>
#	ifdef __WINCE__
#		include <af_irda.h>
#		if defined(WCE_AF_IRDA) && (AF_IRDA != WCE_AF_IRDA)
			// In an MFC app, this is incorrectly set to WINDOWS_AF_IRDA!!!!
#			undef AF_IRDA
#			define AF_IRDA WCE_AF_IRDA
#		endif
#	endif
#else
#	include <sys/types.h>
#	include <sys/socket.h>
#	include <netdb.h>
#	include <netinet/in.h>
#	include <arpa/inet.h>
#	include <unistd.h>
#	define SOCKET 			int
#	define SOCKADDR			struct sockaddr
#	define PSOCKADDR		struct sockaddr*
#	define SOCKADDR_IN		struct sockaddr_in
#	define INVALID_SOCKET 	-1
#	define SOCKET_ERROR		-1
#	define LPHOSTENT		struct hostent*
#	define IN_ADDR			struct in_addr
#	define PIN_ADDR			struct in_addr*
#endif

#define RECVBUF_SIZE	1024 * 16
#define PING_ERROR		0xFFFFFFFF

#define SOCKETRECV_OK		1
#define SOCKETRECV_TIMEOUT	2
#define SOCKETRECV_CLOSED	4
#define SOCKETRECV_ERROR	8

enum socketErrors_t {seNone, seSocket, seTimeOut};

class TSocket {
public:
	TSocket();
	TSocket(const TSocket& copy);
	virtual ~TSocket();

	SOCKET fileno();
	static TString GetHostName();
	static TString GetHostByName(TString hostName);
	static VALUE GetAllHostByName(TString hostName);
	bool CreateSocket(int af = AF_INET, int type = SOCK_STREAM, int protocol = 0);
	virtual bool Connect(LPCTSTR hostName, int port);
	void Close();
	int Send(const char* sendBuf, int sizeSendBuf);
	int Receive(char* recvBuf, int sizeRecvBuf);
	int Read(TString& result);
	int ReadLine(TString& result);
	void Write(TString data);
	void WriteLine(TString data);
	bool Bind(LPCTSTR hostName, int port);
	bool Listen(int backlog);
	TSocket* Accept(TString& whoConnected);
	TString GetIP();

#if defined(__WINNT__) || defined(__WINCE__)
	DWORD Ping(LPCTSTR hostName, float timeoutms);
#endif

	// these two are intended to be Unicode compliant; i.e. they work under CE
	virtual void Send(TString text);
	//virtual TString Receive();
	virtual int Receive(TString& result);

	int GetTimeOut();
	void SetTimeOut(DWORD timeOut = 10000);
	void SetTerminator();							// Don't use a terminator
	void SetTerminator(char termChar);				// Use this as a terminator
	void SetTerminator2(char termChar, char termChar2);
	virtual bool GetTerminator(TString& term);		// Returns true if terminator is set
													// and places the terminator char(s) in term

	virtual void Error(socketErrors_t type, TString message);
	socketErrors_t m_LastErrorType;
	TString m_LastErrorMsg;
	bool WasError();

protected:
	bool GenerateSockAddr();
	TString SockErrorStr(int error);

	DWORD m_TimeOut;	// Time out in milliseconds
	char m_Terminator;	// End of send and end of receive marker
	char m_Terminator2;
	int m_bUseTerminator;
	int m_AF;
	int m_Type;

	int m_RecvCount;
	int m_RecvStart;
	char m_RecvBuf[RECVBUF_SIZE];

#if defined(__WINNT__) || defined(__WINCE__)
	static long m_InstanceCount;
	static WSADATA m_wsaData;
#endif

	TCHAR m_HostName[128];
	int m_Port;
	bool m_bConnected;
	SOCKET m_Socket;
	SOCKADDR_IN m_sockAddr;
#if defined(__WINCE__)
	SOCKADDR_IRDA m_irdaSockAddr;
#endif
};


#endif	// _TSOCKET___H