/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

// DebugOutDlg.cpp : implementation file
//

#include "stdafx.h"
#include "DialectApp.h"

#ifdef IDE_SUPPORT

#include "DebugOutDlg.h"
#include "CValue.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CDebugOutDlg dialog


CDebugOutDlg::CDebugOutDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CDebugOutDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CDebugOutDlg)
	//}}AFX_DATA_INIT
}


void CDebugOutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CDebugOutDlg)
	DDX_Control(pDX, IDC_OUTPUT_EDIT, m_OutputEdit);
	DDX_Control(pDX, IDC_EXECSEL_BTN, m_ExecBtn);
	DDX_Control(pDX, IDC_EVALSEL_BTN, m_EvalBtn);
	DDX_Control(pDX, IDC_CLEAR_BTN, m_ClearBtn);
	DDX_Control(pDX, IDC_DEBUG_EDIT, m_Edit);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CDebugOutDlg, CDialog)
	//{{AFX_MSG_MAP(CDebugOutDlg)
	ON_WM_SIZE()
	ON_BN_CLICKED(IDC_CLEAR_BTN, OnClearBtn)
	ON_BN_CLICKED(IDC_EVALSEL_BTN, OnEvalselBtn)
	ON_BN_CLICKED(IDC_EXECSEL_BTN, OnExecselBtn)
	ON_WM_CLOSE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CDebugOutDlg message handlers

void CDebugOutDlg::OnSize(UINT nType, int cx, int cy)
{
	CDialog::OnSize(nType, cx, cy);
	if (::IsWindow(m_OutputEdit.GetSafeHwnd())) {
		m_OutputEdit.MoveWindow(0, 0, cx, 80);
		m_Edit.MoveWindow(0, 85, cx, cy - 85 - 35);
		m_EvalBtn.MoveWindow(10, cy - 30, 120, 27);
		m_ExecBtn.MoveWindow(140, cy - 30, 120, 27);
		m_ClearBtn.MoveWindow(270, cy - 30, 120, 27);
	}
}

void CDebugOutDlg::OnClearBtn()
{
	m_Edit.SetWindowText(_T(""));
	m_Edit.SetFocus();
}

void CDebugOutDlg::OnEvalselBtn()
{
	VALUE vResult;
	CString exception, code = GetCode();
	if (g_Interpreter.DebugEvaluate(code, vResult, exception)) {
		WriteResult(vResult.String());
	}
	else {
		WriteResult(exception);
	}
	m_Edit.SetFocus();
}

void CDebugOutDlg::OnExecselBtn()
{
	if (!g_Interpreter.InBreakpoint()) {
		MessageBox(_T("Execute selection can only be used when at a breakpoint"));
		return;
	}

	CString exception, code = GetCode();
	if (!g_Interpreter.DebugInterpretText(code, exception)) {
		WriteResult(exception);
	}
	m_Edit.SetFocus();
}

BOOL CDebugOutDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	m_Font.CreatePointFont(100, _T("Courier New"));
	m_Edit.SetFont(&m_Font);

#ifdef UNDER_CE
	RECT rc;
	GetWindowRect(&rc);

	RECT rcScreen;
	SystemParametersInfo(SPI_GETWORKAREA, 0, &rcScreen, 0);

	if (rc.bottom > rcScreen.bottom) {
		rc.top = 0;
		rc.bottom = rcScreen.bottom;
		MoveWindow(&rc);
	}
#endif

	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

CString CDebugOutDlg::GetCode()
{
	CString code;
	int start, end;
	m_Edit.GetSel(start, end);
	if (start == end) {
		// Execute the entire line
		int line = m_Edit.LineFromChar();
		int max = m_Edit.GetWindowTextLength();
		TCHAR* buf = code.GetBuffer(max + 1);
		int copied = m_Edit.GetLine(line, buf, max);
		code.ReleaseBuffer(copied);
	}
	else {
		// Execute just the selection
		CString text;
		m_Edit.GetWindowText(text);
		code = text.Mid(start, end - start);
	}
	return code;
}

void CDebugOutDlg::WriteResult(CString result)
{
	int start, end;
	m_Edit.GetSel(start, end);
	int endLine = m_Edit.LineFromChar(end);
	int lineLength = m_Edit.LineLength(endLine);
	int lineCharIdx = m_Edit.LineIndex(endLine);
	m_Edit.SetSel(lineCharIdx + lineLength, lineCharIdx + lineLength);
	m_Edit.ReplaceSel(_T("\r\n---\r\n") + result);
}

void CDebugOutDlg::AddDebugString(CString text)
{
	// Called by the interpreter debug statement
	int size = m_Edit.GetWindowTextLength();
	m_OutputEdit.SetSel(-1, -1);
	m_OutputEdit.ReplaceSel(text + _T("\r\n"));
}

void CDebugOutDlg::ClearDebugOutput()
{
	m_OutputEdit.SetWindowText(_T(""));
}

void CDebugOutDlg::OnClose()
{
	ShowWindow(SW_HIDE);
	//AfxGetMainWnd()->BringWindowToTop();
	AfxGetMainWnd()->SetFocus();
	//CDialog::OnClose();
}

BOOL CDebugOutDlg::PreTranslateMessage(MSG* pMsg)
{
	if ((pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_ESCAPE)) {
		OnClose();
		return TRUE;
	}
	return CDialog::PreTranslateMessage(pMsg);
}

//////////////////////////////////////////////////////////////
#endif // IDE_SUPPORT
