/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

// MakeExeDlg.cpp : implementation file
//

#include "stdafx.h"
#include "DialectApp.h"

#ifdef IDE_SUPPORT

#include "MakeExeDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CMakeExeDlg dialog


CMakeExeDlg::CMakeExeDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CMakeExeDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CMakeExeDlg)
	m_InitialModule = _T("");
	m_OutputFilename = _T("");
	m_SourceDir = _T("");
	m_InterpStub = _T("");
	//}}AFX_DATA_INIT
}


void CMakeExeDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CMakeExeDlg)
	DDX_Text(pDX, IDC_MODULE_EDIT, m_InitialModule);
	DDX_Text(pDX, IDC_OUTFILE_EDIT, m_OutputFilename);
	DDX_Text(pDX, IDC_SRCDIR_EDIT, m_SourceDir);
	DDX_Text(pDX, IDC_STUB_EDIT, m_InterpStub);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CMakeExeDlg, CDialog)
	//{{AFX_MSG_MAP(CMakeExeDlg)
	ON_BN_CLICKED(IDC_BUTTON1, OnButton1)
	ON_BN_CLICKED(IDC_BROWSEINIT_BTN, OnBrowseinitBtn)
	ON_BN_CLICKED(IDC_BROWSEINTERP_BTN, OnBrowseinterpBtn)
	ON_BN_CLICKED(IDC_BROWSEOUT_BTN, OnBrowseoutBtn)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CMakeExeDlg message handlers

BOOL CMakeExeDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	TCHAR buf[MAX_PATH];
	GetModuleFileName(NULL, buf, MAX_PATH);
	m_InterpStub = buf;

	if (!m_SourceDir.IsEmpty()) {
		if (m_SourceDir[m_SourceDir.GetLength() - 1] != '\\') {
			m_SourceDir += '\\';
		}
	}

	m_OutputFilename = m_InitialModule;
	if (!m_OutputFilename.IsEmpty()) {
		if (m_OutputFilename.Right(2).CompareNoCase(_T(".d")) == 0) {
			m_OutputFilename = m_OutputFilename.Left(m_OutputFilename.GetLength() - 2);
		}
		m_OutputFilename += _T(".exe");
	}

	UpdateData(FALSE);

	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void CMakeExeDlg::OnButton1()
{
	UpdateData(TRUE);

	bool bResult;
	{
		CWaitCursor wc;
		bResult = g_Interpreter.MakeExe(m_OutputFilename, m_InterpStub,
			m_SourceDir, m_InitialModule);
	}

	if (!bResult)
	{
		VALUE v = g_Interpreter.GetException();
		CString error;
		error.Format(_T("Failed to make executable: %s\n\nWould you like to try again?"),
			v[_T("msg")].String());
		g_Interpreter.ClearException();
		if (MessageBox(error, _T("Make Exe..."), MB_YESNO) == IDYES) return;
	}

	CDialog::OnOK();
}

CString _GetFileName(CWnd* pParent, CString filter, CString ext)
{
	CString result;
	TCHAR buf[MAX_PATH];
	memset(buf, 0, MAX_PATH * sizeof(TCHAR));
	CFileDialog dlg(TRUE, ext, _T("*.") + ext, OFN_FILEMUSTEXIST, filter, pParent);
	dlg.m_ofn.lpstrFile = buf;
	dlg.m_ofn.nMaxFile = MAX_PATH;
	if (dlg.DoModal() == IDOK) {
		result = buf;
	}
	return result;
}

void CMakeExeDlg::OnBrowseinitBtn()
{
	CString text = _GetFileName(this, _T("Dialect Scripts (*.d)|*.d||"), _T("d"));
	if (!text.IsEmpty()) {
		m_InitialModule = text;
		int pos = text.ReverseFind('\\');
		if (pos > 0) {
			m_SourceDir = text.Left(pos + 1);
		}
		UpdateData(FALSE);
	}
}

void CMakeExeDlg::OnBrowseinterpBtn()
{
	CString text = _GetFileName(this, _T("Executables (*.exe)|*.exe||"), _T("exe"));
	if (!text.IsEmpty()) {
		m_InterpStub = text;
		UpdateData(FALSE);
	}
}

void CMakeExeDlg::OnBrowseoutBtn()
{
	CString filter = _T("Executables (*.exe)|*.exe||");
	CString ext = _T("exe");

	TCHAR buf[MAX_PATH];
	memset(buf, 0, MAX_PATH * sizeof(TCHAR));
	CFileDialog dlg(TRUE, ext, _T("*.") + ext, OFN_PATHMUSTEXIST | OFN_OVERWRITEPROMPT,
		filter, this);
	dlg.m_ofn.lpstrFile = buf;
	dlg.m_ofn.nMaxFile = MAX_PATH;
	if (dlg.DoModal() == IDOK) {
		m_OutputFilename = buf;
		UpdateData(FALSE);
	}
}

#endif // IDE_SUPPORT
