/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

// QButton.cpp : implementation file
//

#include "stdafx.h"
#include "QButton.h"
#include "CImplementer.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

extern bool FrameToFont(VALUE vFrame, CFont& font); // GUIModule.cpp

CQButton::CQButton(CQWindow* pQWindow, CString label,
				   RECT& rc, int options, bool bFocusRect, CQBitmap* pBitmap,
				   VALUE vFont,
				   VALUE (*pHandler)(CQButton*, int, VALUE), VALUE vSelf)
				   :CButton(), CQBase(vSelf)
{
	m_pHandler = pHandler;

	if ((options & QBS_CHECKBOX) || (options & QBS_RADIOBUTTON)) {
		bFocusRect = true;
		pBitmap = NULL;
	}
	m_bFocusRect = bFocusRect;

	Create(
		label,						// LPCTSTR lpszCaption,
		WS_CHILD |
		WS_VISIBLE |
		WS_TABSTOP |
		(pBitmap ? BS_OWNERDRAW : 0) |
		options,				// DWORD dwStyle,
		rc,						// const RECT& rect,
		pQWindow,					// CWnd* pParentWnd,
		0							// UINT nID
		);

	if (options & QBS_MULTILINE) {
			options |= BS_MULTILINE;
	}
	
	
	if ((vFont.Type() == stFrame) && FrameToFont(vFont, m_Font)) {
		SetFont(&m_Font);
	}
	else {
		SetFont(&g_QLabelFont);
	}

	if (pBitmap != NULL) {
		m_bBitmap = true;
		// Copy the bitmap (I'm sure there is a faster way)
		m_Bitmap = *pBitmap;
	}
}


CQButton::~CQButton()
{
}

BEGIN_MESSAGE_MAP(CQButton, CButton)
	//{{AFX_MSG_MAP(CQButton)
	ON_CONTROL_REFLECT(BN_CLICKED, OnClicked)
	ON_WM_SETFOCUS()
	ON_WM_DESTROY()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CQButton message handlers


void CQButton::OnClicked()
{
	if (m_pHandler)	{
		m_pHandler(this, QW_CLICK, VALUE());
	}
}

void CQButton::OnSetFocus(CWnd* pOldWnd)
{
	if (m_bFocusRect) {
		CButton::OnSetFocus(pOldWnd);
	}
}

void CQButton::OnDestroy()
{
	CButton::OnDestroy();
	if (m_pHandler) m_pHandler(this, QW_DESTROY, VALUE());	// Causes garbage collection
}

void CQButton::DrawBitmap(LPCRECT rect, CDC* pDC)
{
	// Get the bitmap info pointer, & some info
	BITMAP bitmap;
	m_Bitmap.GetBitmap(&bitmap);
	int wd = bitmap.bmWidth;
	int ht = bitmap.bmHeight;

	// Create a DC for the bitmap
	CDC bitmapDC;
	bitmapDC.CreateCompatibleDC(NULL); //compatible with the screen
	bitmapDC.SelectObject(m_Bitmap);

	// Copy the bitmap to the screen
	pDC->BitBlt(
		(rect->left + rect->right - wd)/2,	// left dest
		(rect->top + rect->bottom - ht)/2,	// top dest
		wd,					// bm width
		ht,					// bm height
		&bitmapDC,			// pSrcDC
		0,					// left src
		0,					// top src
		SRCCOPY				// operation
	);
}

void CQButton::DrawItem(LPDRAWITEMSTRUCT lpDrawItemStruct)
{
	// Get the bounding rect and the hDC
	RECT rect = lpDrawItemStruct->rcItem;
	CDC dc;
	dc.Attach(lpDrawItemStruct->hDC);

	// Draw the button bounding box
	CBrush Brush(RGB(192,192,192));
	dc.FillRect(&rect,&Brush);
	if (lpDrawItemStruct->itemState & 1)
	{
		rect.left++;rect.top++;rect.right++;rect.bottom++;
		//if (m_drawFunction) m_drawFunction(&rect,&dc);
		if (m_bBitmap) DrawBitmap(&rect,&dc);
		rect.left--;rect.top--;rect.right--;rect.bottom--;
		dc.Draw3dRect(&rect, RGB(0, 0, 0), RGB(255, 255, 255));
	}
	else
	{
		//if (m_drawFunction) m_drawFunction(&rect,&dc);
		if (m_bBitmap) DrawBitmap(&rect,&dc);
		dc.Draw3dRect(&rect, RGB(255, 255, 255), RGB(0, 0, 0));
	}

}
