/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

// QCalendar.cpp : implementation file
//

#include "stdafx.h"
#include "QCalendar.h"
//#include "XTime.h"
#include <afxdisp.h> // COleDateTime

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/*LPCTSTR intToMonth[] = {
	_T(""),
	_T("January"),
	_T("February"),
	_T("March"),
	_T("April"),
	_T("May"),
	_T("June"),
	_T("July"),
	_T("August"),
	_T("September"),
	_T("October"),
	_T("November"),
	_T("December")
};*/

#if UNDER_CE < 210
extern BOOL __GetAsSystemTime(const COleDateTime& dt, SYSTEMTIME& sysTime); // CoreFunctions.cpp
#endif

/////////////////////////////////////////////////////////////////////////////
// CQCalendar

CQCalendar::CQCalendar(CQWindow* pParent, CString date, RECT& rc,
					   VALUE (*pHandler)(CQCalendar*, int, VALUE), VALUE vSelf)
					   : CWnd(), CQBase(vSelf)
{
	if (!pParent) return;

	m_pHandler = pHandler;

	static bDoneInit = false;
	if (!bDoneInit) {
		bDoneInit = true;
		INITCOMMONCONTROLSEX icc;
		icc.dwSize = sizeof(INITCOMMONCONTROLSEX);
		icc.dwICC = ICC_DATE_CLASSES;
		InitCommonControlsEx(&icc);
	}

	if (!CreateEx(0, MONTHCAL_CLASS, _T(""),
		WS_BORDER | WS_CHILD | WS_VISIBLE,
		rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top,
		pParent->m_hWnd, NULL))
		return;

	// Get the size required to show an entire month and make it so
	RECT actualrc;
	MonthCal_GetMinReqRect(m_hWnd, &actualrc);
	SetWindowPos(NULL, 0, 0, actualrc.right, actualrc.bottom, SWP_NOZORDER | SWP_NOMOVE);

	SetDate(date);
}

CQCalendar::~CQCalendar()
{
}


BEGIN_MESSAGE_MAP(CQCalendar, CWnd)
	//{{AFX_MSG_MAP(CQCalendar)
	ON_WM_DESTROY()
	ON_NOTIFY_REFLECT(MCN_SELCHANGE, OnSelectionChange)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


/////////////////////////////////////////////////////////////////////////////
// CQCalendar message handlers

void CQCalendar::SetDate(CString date)
{
	//CXTime t;
	COleDateTime t;
	SYSTEMTIME st;
	if (date.IsEmpty()) {
		GetLocalTime(&st);
	}
	else {
		_TRY {
#if UNDER_CE < 210
			if (t.ParseDateTime(date) && __GetAsSystemTime(t, st)) {
#else
			if (t.ParseDateTime(date) && t.GetAsSystemTime(st)) {
#endif
			}
			else {
				GetLocalTime(&st);
			}
		}
		_CATCH_ALL {
			GetLocalTime(&st);
		}
	}
	MonthCal_SetCurSel(m_hWnd, &st);
}

void CQCalendar::OnDestroy()
{
	CWnd::OnDestroy();
	if (m_pHandler) m_pHandler(this, QW_DESTROY, VALUE());	// Causes garbage collection
}

CString CQCalendar::ShortDate()
{
	// Like 5/12/1987
	CString date;
	SYSTEMTIME st;
	if (MonthCal_GetCurSel(m_hWnd, &st)) {
		//date.Format(_T("%d/%d/%d"), st.wMonth, st.wDay, st.wYear);
		LPTSTR buf = date.GetBuffer(128);
		int count = GetDateFormat(LOCALE_USER_DEFAULT, DATE_SHORTDATE, &st, NULL, buf, 128);
		date.ReleaseBuffer(count > 0 ? count - 1 : 0);
	}
	return date;
}

CString CQCalendar::LongDate()
{
	// Like May 12, 1997
	CString date;
	SYSTEMTIME st;
	if (MonthCal_GetCurSel(m_hWnd, &st)) {
		//date.Format(_T("%s %d, %d"), intToMonth[st.wMonth], st.wDay, st.wYear);
		LPTSTR buf = date.GetBuffer(128);
		int count = GetDateFormat(LOCALE_USER_DEFAULT, DATE_LONGDATE, &st, NULL, buf, 128);
		date.ReleaseBuffer(count > 0 ? count - 1 : 0);
	}
	return date;
}

BOOL CQCalendar::PreTranslateMessage(MSG* pMsg)
{
	return CWnd::PreTranslateMessage(pMsg);
}

void CQCalendar::OnSelectionChange(NMHDR* pNotifyStruct, LRESULT* result)
{
	*result = TRUE;
	if (m_pHandler) {
		m_pHandler(this, QW_SELCHANGE, VALUE());
	}
}
