/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

// QEdit.cpp : implementation file
//

#include "stdafx.h"
#include "..\DialectLib\CValue.h"
#include "QEdit.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

CQEdit::CQEdit(CQWindow* pQWindow, CString text, RECT& rc, bool bMultiline, VALUE vFont,
			   int options, VALUE (*pHandler)(CQEdit*, int, VALUE), VALUE vSelf)
			   :CEdit(), CQBase(vSelf)
{
	m_pHandler = pHandler;
	m_bMultiLine = bMultiline;
	if (m_bMultiLine) {
		options = options | ES_MULTILINE | ES_WANTRETURN;
	}

	CreateEx(options & QES_BORDER ? WS_EX_CLIENTEDGE : 0,
		_T("EDIT"), NULL,
		WS_CHILD | WS_VISIBLE |	WS_TABSTOP | options,
		rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top,
		pQWindow ? pQWindow->m_hWnd : NULL, 0);

	int dlgUnits(16);
	SetTabStops(dlgUnits);

	if (vFont.Type() == stFrame) {
		int size = 100;
		CString name = _T("Arial");

		VALUE v = vFont[_T("size")];
		if (v.Type() == stInteger) {
			size = v.Integer();
		}

		v = vFont[_T("name")];
		if (v.Type() == stString) {
			name = v.String();
		}

		if (m_Font.CreatePointFont(size, name)) {
			SetFont(&m_Font, FALSE);
		}
		else {
			SetFont(&g_QLabelFont, FALSE);
		}
	}
	else {
		SetFont(&g_QLabelFont, FALSE);
	}

	SetWindowText(text);
}


CQEdit::~CQEdit()
{
}

#ifdef UNDER_CE
#	define ON_WM_RBUTTONUP()
#endif
BEGIN_MESSAGE_MAP(CQEdit, CEdit)
	//{{AFX_MSG_MAP(CQEdit)
	ON_CONTROL_REFLECT(EN_CHANGE, OnChange)
	ON_CONTROL_REFLECT(EN_KILLFOCUS, OnKillfocus)
	ON_CONTROL_REFLECT(EN_SETFOCUS, OnSetfocus)
	ON_CONTROL_REFLECT(EN_UPDATE, OnUpdate)
	ON_WM_DESTROY()
	ON_WM_RBUTTONUP()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CQEdit message handlers

void CQEdit::OnSetfocus()
{
	if (m_pHandler) {
		m_pHandler(this, QW_FOCUS, VALUE());
	}
}

void CQEdit::OnKillfocus()
{
	if (m_pHandler) {
		m_pHandler(this, QW_UNFOCUS, VALUE());
	}
}

void CQEdit::OnChange()
{
	if (m_pHandler) {
		m_pHandler(this, QW_CHANGE, VALUE());
	}
}

void CQEdit::OnUpdate()
{
	if (m_pHandler) {
		m_pHandler(this, QW_UPDATE, VALUE());
	}
}

void CQEdit::OnDestroy()
{
	CEdit::OnDestroy();
	if (m_pHandler) m_pHandler(this, QW_DESTROY, VALUE());	// Causes garbage collection
}

/*void CQEdit::OnLButtonDown(UINT nFlags, CPoint point)
{
	if (m_pHandler) {
		if (m_pHandler(this, QW_CLICK, VALUE()).Type() != stUndefined) return;
	}
	CEdit::OnLButtonDown(nFlags, point);
}*/

#ifndef UNDER_CE
void CQEdit::OnRButtonUp(UINT nFlags, CPoint point)
{
	if (m_pHandler) {
		if (m_pHandler(this, QW_RCLICK, VALUE()).Type() != stUndefined) return;
	}
	CEdit::OnRButtonUp(nFlags, point);
}
#endif

BOOL CQEdit::PreTranslateMessage(MSG* pMsg)
{
#ifndef UNDER_CE
	// Hack around an IsDialogMessage() problem in QWindow
	if ((pMsg->message == WM_KEYDOWN) && (pMsg->wParam == VK_TAB)) {
		if (m_bMultiLine) {
			ReplaceSel(_T("\t"), TRUE);
			return TRUE;
		}
	}
	else if (((pMsg->message == WM_KEYDOWN) || (pMsg->message == WM_KEYUP) || (pMsg->message == WM_CHAR)) &&
		(pMsg->wParam == VK_ESCAPE)) {
		return 1;
	}
	else if ((pMsg->message == WM_LBUTTONUP) && m_pHandler) {
		if (m_pHandler(this, QW_CLICK, VALUE()).Type() != stUndefined)
			return 1;
	}
#else
	// Hack around OnRButton MFC macro
	if ((pMsg->message == WM_RBUTTONUP) && m_pHandler) {
		if (m_pHandler(this, QW_RCLICK, VALUE()).Type() != stUndefined)
			return 0;
	}
#endif
	return CEdit::PreTranslateMessage(pMsg);
}
