/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

// QInk.cpp : implementation file
//

#include "stdafx.h"
#include "QInk.h"
#include "CImplementer.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

CQInk::CQInk(CQWindow* pQWindow, RECT& rc, BorderStyle borderStyle, VALUE vData,
		int options, VALUE (*pHandler)(CQInk*, int, VALUE), VALUE vSelf)
		:CInkWnd(), CQBase(vSelf)
{
	Create(pQWindow, vData, rc, borderStyle);
	m_pHandler = pHandler;
}

CQInk::~CQInk(){}

BEGIN_MESSAGE_MAP(CQInk, CInkWnd)
	//{{AFX_MSG_MAP(CQInk)
	ON_WM_DESTROY()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

void CQInk::OnDestroy()
{
	CInkWnd::OnDestroy();
	if (m_pHandler) m_pHandler(this, QW_DESTROY, VALUE());	// Causes garbage collection
}

/////////////////////////////////////////////////////////////////////////////
// CInkWnd

CInkWnd::CInkWnd():CWnd()
{
	m_nPoint = 0;
}

bool CInkWnd::IsEmpty()
{
	return m_nPoint == 0;
}

BOOL CInkWnd::Create(CWnd* pParent, VALUE vInitData, RECT& rc, BorderStyle borderStyle)
{
	m_BorderStyle = borderStyle;
	if (CWnd::Create(NULL, NULL, WS_VISIBLE | WS_CHILD, rc, pParent, 0)) {
		setData(vInitData);
		return TRUE;
	}
	return FALSE;
}

CInkWnd::~CInkWnd()
{
}

BEGIN_MESSAGE_MAP(CInkWnd, CWnd)
	//{{AFX_MSG_MAP(CInkWnd)
	ON_WM_MOUSEMOVE()
	ON_WM_PAINT()
	ON_WM_SIZE()
	ON_WM_LBUTTONUP()
	ON_WM_LBUTTONDOWN()
	ON_WM_KILLFOCUS()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


/////////////////////////////////////////////////////////////////////////////
// CInkWnd message handlers

void CInkWnd::SetBorderStyle(BorderStyle b)
{
	if (b != m_BorderStyle) {
		m_BorderStyle = b;
		InvalidateRect(NULL);
	}
}

CInkWnd::BorderStyle CInkWnd::GetBorderStyle()
{
	return m_BorderStyle;
}

void CInkWnd::OnPaint()
{
	CPaintDC dc(this); // device context for painting

	RECT rc;
	GetClientRect(&rc);
	dc.FillSolidRect(&rc, RGB(255, 255, 255));

	int i = 0;
	CPoint pZero(0,0);
	while(i < m_nPoint) {
		int k = 0, m = i;
		while((m_Points[i] != pZero) && (i < m_nPoint)) {
			i++;
			k++;
		}
		i++;
		dc.Polyline(&(m_Points[m]), k);
	}

	if (m_BorderStyle == bsRaised) {
		dc.Draw3dRect(&rc, RGB(255, 255, 255), RGB(0, 0, 0));
	}
	else if (m_BorderStyle == bsSunken) {
		dc.Draw3dRect(&rc, RGB(0, 0, 0),  RGB(255, 255, 255));
	}
	else if (m_BorderStyle == bsLine) {
		CBrush brush;
		brush.CreateSolidBrush(RGB(0, 0, 0));
		CBrush* pOldBrush = dc.SelectObject(&brush);
		dc.FrameRect(&rc, &brush);
		dc.SelectObject(pOldBrush);
	}
}

void CInkWnd::OnSize(UINT nType, int cx, int cy)
{
	CWnd::OnSize(nType, cx, cy);
}

void CInkWnd::OnMouseMove(UINT nFlags, CPoint point)
{
	if ((nFlags & MK_LBUTTON) && (m_nPoint < MAX_STROKE_POINTS - 1)) {
		CDC* pDC = GetDC();
		m_Line[1] = point;
		if (m_Line[0] != m_Line[1]) {
			pDC->Polyline(m_Line, 2);
			m_Points[m_nPoint++] = point;
			//if (m_nPoint >= MAX_STROKE_POINTS) {
			//	m_nPoint = 0;
			//}
		}
		m_Line[0] = m_Line[1];
		ReleaseDC(pDC);
	}
}

void CInkWnd::OnLButtonDown(UINT nFlags, CPoint point)
{
	if (m_nPoint < MAX_STROKE_POINTS - 1) {
		m_Line[0] = point;
		m_Points[m_nPoint++] = point;
		//if (m_nPoint >= MAX_STROKE_POINTS) {
		//	m_nPoint = 0;
		//}
	}
}

void CInkWnd::OnLButtonUp(UINT nFlags, CPoint point)
{
	if (m_nPoint < MAX_STROKE_POINTS) {
		CPoint pointTmp(0,0);
		m_Points[m_nPoint++] = pointTmp;
		//if (m_nPoint >= MAX_STROKE_POINTS) {
		//	m_nPoint = 0;
		//}
	}
}

void CInkWnd::OnKillFocus(CWnd* pNewWnd)
{
	if (m_nPoint < MAX_STROKE_POINTS) {
		CPoint pointTmp(0,0);
		m_Points[m_nPoint++] = pointTmp;
		//if (m_nPoint >= MAX_STROKE_POINTS) {
		//	m_nPoint = 0;
		//}
	}
}

VALUE CInkWnd::getData()
{
	RECT rc;
	GetClientRect(&rc);

	int count = m_nPoint;
	short int *pbits = new short int[count * 2 + 4];
	short int *head = pbits;
	*pbits++ = 1;			// Structure Version
	*pbits++ = rc.right;	// Width
	*pbits++ = rc.bottom;	// Height
	*pbits++ = count;		// Point Count
	for (int idx = 0; idx < count; ++idx) {
		*pbits++ = m_Points[idx].x;
		*pbits++ = m_Points[idx].y;
	}
	VALUE result = AllocChunk((count * 2 + 4) * sizeof(short int), (BYTE*)head);
	delete[] head;
	return result;
}

bool CInkWnd::setData(VALUE vData)
{
	bool bResult = true;
	m_nPoint = 0;
	if ((vData.Type() == stChunk) && (vData.GetChunkByteCount() >= (sizeof(int) * 4))) {
		m_nPoint = 0;
		short int* pbits = (short int*)vData.GetChunkData();
		int version = (int)*pbits++;
		int width = (int)*pbits++;
		int height = (int)*pbits++;
		int count = (int)*pbits++;
		for (m_nPoint = 0; m_nPoint < count; ++m_nPoint) {
			if (m_nPoint >= MAX_STROKE_POINTS) break;
			int x = (int)*pbits++;
			int y = (int)*pbits++;
			m_Points[m_nPoint] = CPoint(x, y);
		}
	}
	else if (vData.Type() != stNil) bResult = false;

	InvalidateRect(NULL);
	return bResult;
}
