/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

// QLabel.cpp : implementation file
//

#include "stdafx.h"
#include "QLabel.h"
#include "QBitmap.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

extern CFont g_QLabelFont;
extern bool FrameToFont(VALUE vFrame, CFont& font); // GUIModule.cpp

/////////////////////////////////////////////////////////////////////////////
// CQLabel

CQLabel::CQLabel(CQWindow* pQWindow, CString label, RECT& rc, int options,
				 CQBitmap* pBitmap, VALUE vFont,
				 VALUE (*pHandler) (CQLabel*, int, VALUE), VALUE vSelf)
				 : CStatic(), CQBase(vSelf)
{
	m_pHandler = pHandler;

	if (pBitmap) {
		m_Bitmap = *pBitmap;
		if (options & SS_CENTER) {
			options &= ~SS_CENTER;
			options |= SS_CENTERIMAGE;
		}
		Create(_T(""), WS_CHILD | WS_VISIBLE | SS_NOTIFY | SS_BITMAP | options,
			rc, pQWindow);
		SetBitmap(m_Bitmap);
	}
	else {
		Create(label, WS_CHILD | WS_VISIBLE | SS_NOTIFY | options, rc, pQWindow);
	}

	if ((vFont.Type() == stFrame) && FrameToFont(vFont, m_Font)) {
		SetFont(&m_Font);
	}
	else {
		SetFont(&g_QLabelFont);
	}
}

CQLabel::~CQLabel()
{
}


BEGIN_MESSAGE_MAP(CQLabel, CStatic)
	//{{AFX_MSG_MAP(CQLabel)
	ON_WM_LBUTTONUP()
	ON_WM_DESTROY()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CQLabel message handlers

void CQLabel::OnLButtonUp(UINT nFlags, CPoint point)
{
	if (m_pHandler) {
		m_pHandler(this, QW_CLICK, VALUE());
	}
}

BOOL CQLabel::PreTranslateMessage(MSG* pMsg)
{
	if (pMsg->message == WM_LBUTTONUP) {
		if (m_pHandler) {
			m_pHandler(this, QW_CLICK, VALUE());
			return TRUE;
		}
	}

	return CStatic::PreTranslateMessage(pMsg);
}

void CQLabel::SetText(CString text)
{
	DWORD style = GetWindowLong(m_hWnd, GWL_STYLE);
	if (style & SS_BITMAP) {
		SetWindowLong(m_hWnd, GWL_STYLE, style & ~SS_BITMAP);
		SetBitmap(NULL);
		m_Bitmap.DeleteObject();
		SetFont(&g_QLabelFont);
	}
	SetWindowText(text);
}

void CQLabel::SetNewBitmap(CQBitmap* pBitmap)
{
	if (!::IsWindow(m_hWnd)) return;

	DWORD style = GetWindowLong(m_hWnd, GWL_STYLE);
	if (pBitmap) {
		m_Bitmap = *pBitmap; // Copy the bitmap resource
		SetWindowLong(m_hWnd, GWL_STYLE, style | SS_BITMAP);
		SetBitmap(m_Bitmap);
	}
	else {
		SetWindowLong(m_hWnd, GWL_STYLE, style & ~SS_BITMAP);
		SetBitmap(NULL);
		m_Bitmap.DeleteObject();
		SetFont(&g_QLabelFont);
	}
	InvalidateRect(NULL);
	UpdateWindow();
}

CString CQLabel::GetText()
{
	CString result;
	GetWindowText(result);
	return result;
}

void CQLabel::OnDestroy()
{
	CStatic::OnDestroy();
	if (m_pHandler) m_pHandler(this, QW_DESTROY, VALUE());	// Causes garbage collection
}

int CQLabel::TextWidth(CString text)
{
	CGdiObject* pOld;
	CDC* pDC = GetDC();
	if (m_Font.GetSafeHandle()) {
		pOld = pDC->SelectObject(&m_Font);
	}
	else {
		pOld = pDC->SelectObject(&g_QLabelFont);
	}
	CSize extent = pDC->GetTextExtent(text);
	pDC->SelectObject(pOld);
	ReleaseDC(pDC);
	return extent.cx;
}

int CQLabel::TextHeight(CString text)
{
	CGdiObject* pOld;
	CDC* pDC = GetDC();
	if (m_Font.GetSafeHandle()) {
		pOld = pDC->SelectObject(&m_Font);
	}
	else {
		pOld = pDC->SelectObject(&g_QLabelFont);
	}
	CSize extent = pDC->GetTextExtent(text);
	pDC->SelectObject(pOld);
	ReleaseDC(pDC);
	return extent.cy;
}
