/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

// CQLayout.cpp : implementation file
//

#include "stdafx.h"
#include "QLayout.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

CQLayoutChild::CQLayoutChild(CQLayout* pParent, VALUE vLayout, VALUE vData,
							 VALUE vRuleContext, bool bButtons, bool bSingleSect, bool bNoFormName)
:CLayoutWindow(pParent, vLayout, vData, vRuleContext, bButtons, bSingleSect, bNoFormName)
{
	m_pParent = pParent;
}

void CQLayoutChild::OnItemInfo(CString itemType, CString dataName, CString listName,
							   VALUE vData, VALUE vItemOptions)
{
	if (m_pParent && m_pParent->OnItemInfo(itemType, dataName, listName, vData, vItemOptions)) return;
	else CLayoutWindow::OnItemInfo(itemType, dataName, listName, vData, vItemOptions);
}

/////////////////////////////////////////////////////////////////////////////
// CQLayout

static void Layout_ExitHandler(int k, CString entry)
{
	CQLayout* p = (CQLayout*)k;
	if (p != NULL) {
		p->OnCommand(entry);
	}
}

CQLayout::CQLayout(CQWindow* pQWindow, RECT& rc, VALUE vLayout, VALUE vData,
				   VALUE vRuleContext, int options,
				   VALUE (*pHandler) (CQLayout*, int, VALUE), VALUE vSelf)
				   :CQBase(vSelf)
{
	m_pLayoutChild = NULL;
	m_pHandler = pHandler;
	m_pLayoutChild = NULL;	// Must do this before Create because of WM_SIZE

	// First create the backing window (parent for layout)
	if (!Create(NULL, NULL, WS_VISIBLE | WS_CHILD, rc, pQWindow, 0)) {
		AfxMessageBox(_T("Failed to create layout backing window."));
		return;
	}

	// Now create the layout itself
	m_pLayoutChild = new CQLayoutChild(this, vLayout, vData, vRuleContext,
		options & QLYOS_BUTTONS ? true : false,
		options & QLYOS_SINGLESECT ? true : false,
		options & QLYOS_NOFORMNAME ? true : false);
	if (m_pLayoutChild == NULL) {
		AfxMessageBox(_T("Failed to create layout child window."));
		return;
	}
	m_pLayoutChild->SetExitHandler(Layout_ExitHandler, (int)this);
	m_pLayoutChild->ShowWindow(SW_SHOW);
}

CQLayout::~CQLayout()
{
	delete m_pLayoutChild;
}


BEGIN_MESSAGE_MAP(CQLayout, CWnd)
	//{{AFX_MSG_MAP(CQLayout)
	ON_WM_DESTROY()
	ON_WM_SIZE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


/////////////////////////////////////////////////////////////////////////////
// CQLayout message handlers

void CQLayout::OnCommand(VALUE vCommand)
{
	if (m_pHandler) {
		m_pHandler(this, QW_MENUSELECT, vCommand);
	}
}

void CQLayout::SetData(VALUE vData)
{
	m_pLayoutChild->SetData(vData);
}

VALUE CQLayout::GetData()
{
	return m_pLayoutChild->GetData();
}

void CQLayout::SetRuleContext(VALUE vFrame)
{
	m_pLayoutChild->SetRuleContext(vFrame);
}

VALUE CQLayout::GetRuleContext()
{
	return m_pLayoutChild->GetRuleContext();
}

void CQLayout::Update()
{
	m_pLayoutChild->Update();
}

void CQLayout::SetFlashMenu(VALUE menu)
{
	m_pLayoutChild->SetFlashMenu(menu);
}

VALUE CQLayout::GetFlashMenu()
{
	return m_pLayoutChild->GetFlashMenu();
}

bool CQLayout::GetSingleSection()
{
	return m_pLayoutChild->GetSingleSection();
}

bool CQLayout::SetSingleSection(bool bSingle)
{
	return m_pLayoutChild->SetSingleSection(bSingle);
}

void CQLayout::OnDestroy()
{
	CWnd::OnDestroy();
	if (m_pHandler) m_pHandler(this, QW_DESTROY, VALUE());	// Causes garbage collection
}

bool CQLayout::OnItemInfo(CString itemType, CString dataName, CString listName,
							   VALUE vData, VALUE vItemOptions)
{
	// Return true if we process this
	if (m_pHandler) {
		VALUE args = AllocFrame();
		args.Set(_T("itemdesc"), itemType);
		args.Set(_T("dataname"), dataName);
		args.Set(_T("listname"), listName);
		args.Set(_T("options"), vItemOptions);
		if (vData.Type() != stUndefined) args.Set(_T("data"), vData);

		return (m_pHandler(this, QW_ITEMINFO, args).Type() != stUndefined);
	}
	return false;
}

void CQLayout::OnSize(UINT nType, int cx, int cy)
{
	CWnd::OnSize(nType, cx, cy);

	if (m_pLayoutChild && ::IsWindow(m_pLayoutChild->m_hWnd)) {
		RECT rc;
		GetClientRect(&rc);
		m_pLayoutChild->MoveWindow(&rc);
	}
}
