/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

// QScrollBar.cpp : implementation file
//

#include "stdafx.h"
#include "QScrollBar.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CQScrollBar

CQScrollBar::CQScrollBar(CQWindow* pQWindow, bool bHorz, RECT& rc,
						 int rMin, int rMax, int pageSize, int lineStep, int pageStep,
						 VALUE (*pHandler)(CQScrollBar*, int, VALUE), VALUE vSelf)
						 :CQBase(vSelf)
{
	m_pHandler = pHandler;
	Create(WS_CHILD | WS_VISIBLE | (bHorz ? SBS_HORZ : SBS_VERT), rc, pQWindow, 0);

	m_LineStep = lineStep;
	m_PageStep = pageStep;

	SCROLLINFO si;
	si.cbSize = sizeof(SCROLLINFO);
	si.fMask = SIF_PAGE | SIF_RANGE;
	si.nMin = rMin;
	si.nMax = rMax;
	si.nPage = pageSize;
	SetScrollInfo(&si);
}

CQScrollBar::~CQScrollBar()
{
}


BEGIN_MESSAGE_MAP(CQScrollBar, CScrollBar)
	//{{AFX_MSG_MAP(CQScrollBar)
	ON_WM_HSCROLL_REFLECT()
	ON_WM_VSCROLL_REFLECT()
	ON_WM_DESTROY()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CQScrollBar message handlers

void CQScrollBar::HScroll(UINT nSBCode, UINT nPos)
{
	switch (nSBCode) {
		case SB_LINELEFT:
			SetScrollPos(GetScrollPos() - m_LineStep);
			break;
		case SB_LINERIGHT:
			SetScrollPos(GetScrollPos() + m_LineStep);
			break;
		case SB_PAGELEFT:
			SetScrollPos(GetScrollPos() - m_PageStep);
			break;
		case SB_PAGERIGHT:
			SetScrollPos(GetScrollPos() + m_PageStep);
			break;
		case SB_THUMBTRACK:
		case SB_THUMBPOSITION:
			SetScrollPos(nPos & 0xFFFF);
			break;
		case SB_TOP:
			break;
		case SB_BOTTOM:
			break;
	}

	if (m_pHandler) {
		m_pHandler(this, QW_HSCROLL, VALUE());
	}
}

void CQScrollBar::VScroll(UINT nSBCode, UINT nPos)
{
	CString temp;
	switch (nSBCode) {
		case SB_LINEUP:
			SetScrollPos(GetScrollPos() - m_LineStep);
			break;
		case SB_LINEDOWN:
			SetScrollPos(GetScrollPos() + m_LineStep);
			break;
		case SB_PAGEUP:
			SetScrollPos(GetScrollPos() - m_PageStep);
			break;
		case SB_PAGEDOWN:
			SetScrollPos(GetScrollPos() + m_PageStep);
			break;
		case SB_THUMBTRACK:
		case SB_THUMBPOSITION:
			SetScrollPos(nPos & 0xFFFF);
			break;
		case SB_TOP:
			break;
		case SB_BOTTOM:
			break;
	}

	if (m_pHandler) {
		m_pHandler(this, QW_VSCROLL, VALUE());
	}
}

void CQScrollBar::OnDestroy()
{
	CScrollBar::OnDestroy();
	if (m_pHandler) m_pHandler(this, QW_DESTROY, VALUE());	// Causes garbage collection
}
