/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

// QListBox.cpp : implementation file
//

#include "stdafx.h"
#include "QListBox.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CQListBox

CQListBox::CQListBox(CQWindow* pQWindow, RECT& rc, int options,
			   VALUE (*pHandler)(CQListBox*, int, VALUE), VALUE vSelf)
			   :CListBox(), CQBase(vSelf)
{
	m_pHandler = pHandler;

	CreateEx(options & QLBS_BORDER ? WS_EX_CLIENTEDGE : 0,
		_T("LISTBOX"), NULL,
		WS_CHILD | WS_VISIBLE |	WS_TABSTOP | WS_HSCROLL | WS_VSCROLL |
		LBS_HASSTRINGS | LBS_NOTIFY | LBS_NOINTEGRALHEIGHT | options,
		rc.left, rc.top, rc.right - rc.left, rc.bottom - rc.top,
		pQWindow->m_hWnd, NULL);

	SetFont(&g_QLabelFont);
}

CQListBox::~CQListBox()
{
}

#ifdef UNDER_CE
#	define 	ON_WM_RBUTTONUP()
#endif
BEGIN_MESSAGE_MAP(CQListBox, CListBox)
	//{{AFX_MSG_MAP(CQListBox)
	ON_CONTROL_REFLECT(LBN_SETFOCUS, OnSetfocus)
	ON_CONTROL_REFLECT(LBN_KILLFOCUS, OnKillfocus)
	ON_CONTROL_REFLECT(LBN_DBLCLK, OnDblclk)
	ON_CONTROL_REFLECT(LBN_SELCHANGE, OnSelchange)
	ON_CONTROL_REFLECT(LBN_SELCANCEL, OnSelcancel)
	ON_WM_DESTROY()
	ON_WM_RBUTTONUP()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CQListBox message handlers

void CQListBox::OnSetfocus()
{
	if (m_pHandler) {
		m_pHandler(this, QW_FOCUS, VALUE());
	}
}

void CQListBox::OnKillfocus()
{
	if (m_pHandler) {
		m_pHandler(this, QW_UNFOCUS, VALUE());
	}
}

void CQListBox::OnDblclk()
{
	if (m_pHandler) {
		m_pHandler(this, QW_DOUBLECLICK, VALUE());
	}
}

void CQListBox::OnSelchange()
{
	if (m_pHandler) {
		m_pHandler(this, QW_SELCHANGE, VALUE());
	}
}

void CQListBox::OnSelcancel()
{
}

void CQListBox::OnDestroy()
{
	CListBox::OnDestroy();
	if (m_pHandler) m_pHandler(this, QW_DESTROY, VALUE());	// Causes garbage collection
}

void CQListBox::OnRButtonUp(UINT nFlags, CPoint point)
{
	// Force the selection to move to item right clicked
	// which is not default behaviour for Windows list boxes
	int itemHeight = GetItemHeight(0);
	int top = GetTopIndex();
	int item = top + (point.y / itemHeight);
	SetCurSel(item);
	OnSelchange();	// Not called automaticly by SetCurSel()

	if (m_pHandler) {
		m_pHandler(this, QW_RCLICK, VALUE());
	}
}

BOOL CQListBox::PreTranslateMessage(MSG* pMsg)
{
#ifdef UNDER_CE
	// Currently needed MFC macro hack
	if (pMsg->message == WM_RBUTTONUP) {
		OnRButtonUp(pMsg->wParam, CPoint(LOWORD(pMsg->lParam), HIWORD(pMsg->lParam)));
		return 0;
	}
#endif
	return CListBox::PreTranslateMessage(pMsg);
}
