/*
 *
 *
Copyright (c) 1998-2001 Aristar, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. The end-user documentation included with the redistribution, if any,
must include the following acknowledgment:

"This product includes software developed by Aristar, Inc. (http://www.aristar.com/)."

Alternately, this acknowledgment may appear in the software itself, if and
wherever such third-party acknowledgments normally appear.

4. The names "Dialect" and "Aristar, Inc." must not be used to endorse or
promote products derived from this software without prior written permission.
For written permission, please contact info@aristar.com.

5. Products derived from this software may not be called "Dialect", nor may
"Dialect" appear in their name, without prior written permission of Aristar, Inc.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
ARISTAR, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This software consists of voluntary contributions made by many individuals
on behalf of Aristar, Inc.  For more information on Aristar, Inc.,
please see <http://www.aristar.com/>.

Portions of this software are based upon software freely available and
written by individuals not associated with Aristar, Inc.  Copyrights are
maintained in the relevant source code files.
 *
 *
 */

// QPane.cpp : implementation file
//

#include "stdafx.h"
#include "QPane.h"
#include "CImplementer.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CQPane

CQPane::CQPane(CQWindow* pQWindow, CString text, RECT& rc, BorderStyle borderStyle,
		int options, VALUE (*pHandler)(CQPane*, int, VALUE), VALUE vSelf)
		:CWnd(), CQBase(vSelf)
{
	m_pPaintDC = NULL;
	SetPen(1, RGB(0, 0, 0));
	SetFont(_T("Arial"), 100);
	Create(NULL, NULL, WS_VISIBLE | WS_CHILD | WS_TABSTOP, rc, pQWindow, 0);
	m_pHandler = pHandler;
	m_BorderStyle = borderStyle;
}

CQPane::~CQPane()
{
}

#if defined(UNDER_CE) && (UNDER_CE <= 200)
#	define 	ON_WM_MBUTTONUP()
#	define 	ON_WM_MBUTTONDOWN()
#	define 	ON_WM_RBUTTONUP()
#	define 	ON_WM_RBUTTONDOWN()
#endif
BEGIN_MESSAGE_MAP(CQPane, CWnd)
	//{{AFX_MSG_MAP(CQPane)
	ON_WM_DESTROY()
	ON_WM_SHOWWINDOW()
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	ON_WM_PAINT()
	ON_WM_SIZE()
	ON_WM_LBUTTONDOWN()
	ON_WM_MBUTTONDOWN()
	ON_WM_MBUTTONUP()
	ON_WM_RBUTTONDOWN()
	ON_WM_RBUTTONUP()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


/////////////////////////////////////////////////////////////////////////////
// CQPane message handlers

void CQPane::SetBorderStyle(BorderStyle b)
{
	if (b != m_BorderStyle) {
		m_BorderStyle = b;
		InvalidateRect(NULL);
	}
}

CQPane::BorderStyle CQPane::GetBorderStyle()
{
	return m_BorderStyle;
}

void CQPane::OnDestroy()
{
	CWnd::OnDestroy();
	if (m_pHandler) m_pHandler(this, QW_DESTROY, VALUE());	// Causes garbage collection
}

BOOL CQPane::PreTranslateMessage(MSG* pMsg)
{
#ifdef UNDER_CE
	// Currently needed MFC macro hack
	if (pMsg->message == WM_RBUTTONUP) {
		OnRButtonUp(pMsg->wParam, CPoint(LOWORD(pMsg->lParam), HIWORD(pMsg->lParam)));
		return 0;
	}
	else if (pMsg->message == WM_RBUTTONDOWN) {
		OnRButtonDown(pMsg->wParam, CPoint(LOWORD(pMsg->lParam), HIWORD(pMsg->lParam)));
		return 0;
	}
	else if (pMsg->message == WM_MBUTTONUP) {
		OnMButtonUp(pMsg->wParam, CPoint(LOWORD(pMsg->lParam), HIWORD(pMsg->lParam)));
		return 0;
	}
	else if (pMsg->message == WM_MBUTTONDOWN) {
		OnMButtonDown(pMsg->wParam, CPoint(LOWORD(pMsg->lParam), HIWORD(pMsg->lParam)));
		return 0;
	}
	return CWnd::PreTranslateMessage(pMsg);
#else
	// Allow tabbing between window items...
	if(IsDialogMessage(pMsg)) {
		return TRUE;
	}
	else {
		return CWnd::PreTranslateMessage(pMsg);
	}
#endif
}

void CQPane::OnPaint()
{
	CPaintDC dc(this); // device context for painting

	RECT rc;
	GetClientRect(&rc);
	dc.FillSolidRect(&rc, GetSysColor(COLOR_BTNFACE));	// remove to keep pane from auto clearing

	if (m_pHandler) {
		m_pPaintDC = &dc;	// So we can find it while in m_pHandler
		m_pHandler(this, QW_PAINT, VALUE());
		m_pPaintDC = NULL;
	}

	if (m_BorderStyle == bsRaised) {
		dc.Draw3dRect(&rc, RGB(255, 255, 255), RGB(0, 0, 0));
	}
	else if (m_BorderStyle == bsSunken) {
		dc.Draw3dRect(&rc, RGB(0, 0, 0),  RGB(255, 255, 255));
	}
	else if (m_BorderStyle == bsLine) {
		CBrush brush;
		brush.CreateSolidBrush(RGB(0, 0, 0));
		CBrush* pOldBrush = dc.SelectObject(&brush);
		dc.FrameRect(&rc, &brush);
		dc.SelectObject(pOldBrush);
	}
}

void CQPane::OnShowWindow(BOOL bShow, UINT nStatus)
{
	CWnd::OnShowWindow(bShow, nStatus);
	if (m_pHandler && m_pCI && (m_pCI->__get__(_T("#pQ")).Type() == stVoid)) {
		m_pHandler(this, QW_VISIBLE, (bool)(bShow == TRUE));
	}
}

void CQPane::OnSize(UINT nType, int cx, int cy)
{
	CWnd::OnSize(nType, cx, cy);
	if (m_pHandler && m_pCI && (m_pCI->__get__(_T("#pQ")).Type() == stVoid)) {
		m_pHandler(this, QW_RESIZE, VALUE());
	}
}

bool CQPane::SetPen(int width, COLORREF color)
{
	m_PenColor = color;
	m_PenWidth = width;
	m_Pen.DeleteObject();
	return m_Pen.CreatePen(PS_SOLID, m_PenWidth, m_PenColor) != 0;
}

void CQPane::Line(int x1, int y1, int x2, int y2)
{
	if (m_pPaintDC != NULL) {
		CPen *oldPen = m_pPaintDC->SelectObject(&m_Pen);
		m_pPaintDC->MoveTo(x1, y1);
		m_pPaintDC->LineTo(x2, y2);
		m_pPaintDC->SelectObject(oldPen);
	}
}

void CQPane::DrawPoint(int x, int y, const COLORREF *pcolor)
{
	if (m_pPaintDC != NULL) {
		m_pPaintDC->SetPixel(x, y, pcolor ? *pcolor : m_PenColor);
	}
}

void CQPane::SetTextColor(const COLORREF *pcolor)
{
	if (m_pPaintDC != NULL) {
		m_pPaintDC->SetTextColor(pcolor ? *pcolor : m_PenColor);
		m_pPaintDC->SetBkMode(1);
	}
}

bool CQPane::SetFont(CString fontName, int size)
{
	m_FontName = fontName;
	m_FontSize = size;
	m_Font.DeleteObject();
	return m_Font.CreatePointFont(size, fontName) != 0;
}

void CQPane::DrawText(int x, int y, CString text)
{
	if (m_pPaintDC != NULL) {
		CFont *oldFont = m_pPaintDC->SelectObject(&m_Font);
		m_pPaintDC->ExtTextOut(x, y, 0, NULL, text, NULL);
		m_pPaintDC->SelectObject(oldFont);
	}
}

VALUE MouseData(UINT nFlags, CPoint point)
{
	VALUE v = AllocFrame();
	v.Set(_T("left"), (bool)(nFlags & MK_LBUTTON));
	v.Set(_T("middle"), (bool)(nFlags & MK_MBUTTON));
	v.Set(_T("right"), (bool)(nFlags & MK_RBUTTON));
	v.Set(_T("shift"), (bool)(nFlags & MK_SHIFT));
	v.Set(_T("control"), (bool)(nFlags & MK_CONTROL));
	VALUE a = AllocArray(2);
	a.Set(1, point.x);
	a.Set(2, point.y);
	v.Set(_T("pos"), a);
	return v;
}

void CQPane::OnMouseMove(UINT nFlags, CPoint point)
{
	if (m_pHandler)	{
		m_pHandler(this, QW_MOUSEMOVE, VALUE());
	}
}

void CQPane::OnLButtonDown(UINT nFlags, CPoint point)
{
	if (m_pHandler)	{
		m_pHandler(this, QW_BUTTONDOWN, MouseData(nFlags, point));
	}
}

void CQPane::OnLButtonUp(UINT nFlags, CPoint point)
{
	if (m_pHandler)	{
		m_pHandler(this, QW_BUTTONUP, MouseData(nFlags, point));
		m_pHandler(this, QW_CLICK, VALUE());
	}
}

void CQPane::OnMButtonDown(UINT nFlags, CPoint point)
{
	if (m_pHandler)	{
		m_pHandler(this, QW_BUTTONDOWN, MouseData(nFlags, point));
	}
}

void CQPane::OnMButtonUp(UINT nFlags, CPoint point)
{
	if (m_pHandler)	{
		m_pHandler(this, QW_BUTTONUP, MouseData(nFlags, point));
		m_pHandler(this, QW_CLICK, VALUE());
	}
}

void CQPane::OnRButtonDown(UINT nFlags, CPoint point)
{
	if (m_pHandler)	{
		m_pHandler(this, QW_BUTTONDOWN, MouseData(nFlags, point));
	}
}

void CQPane::OnRButtonUp(UINT nFlags, CPoint point)
{
	if (m_pHandler)	{
		m_pHandler(this, QW_BUTTONUP, MouseData(nFlags, point));
		m_pHandler(this, QW_CLICK, VALUE());
	}
}
